import asyncio
from collections.abc import Iterator
from typing import Annotated, Any

import litellm
import pytest
from litellm.integrations.custom_logger import CustomLogger
from pydantic import AfterValidator, BaseModel

from magentic.chat_model.base import ToolSchemaParseError
from magentic.chat_model.litellm_chat_model import LitellmChatModel
from magentic.chat_model.message import UserMessage
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        ("List three fruits", [list[str]], list),
    ],
)
@pytest.mark.litellm_openai
def test_litellm_chat_model_complete_openai(prompt, output_types, expected_output_type):
    chat_model = LitellmChatModel("gpt-4o")
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.fixture
def litellm_success_callback_calls() -> Iterator[list[dict[str, Any]]]:
    """A list of calls to the `litellm.success_callback`"""
    original_success_callback = litellm.success_callback.copy()
    callback_calls: list[dict[str, Any]] = []

    def _add_call_to_list(kwargs, completion_response, start_time, end_time):
        callback_calls.append({"kwargs": kwargs})

    litellm.success_callback = [_add_call_to_list]
    yield callback_calls
    litellm.success_callback = original_success_callback


@pytest.mark.litellm_openai
def test_litellm_chat_model_metadata(litellm_success_callback_calls):
    """Test that provided metadata is passed to the litellm success callback."""
    chat_model = LitellmChatModel("gpt-4o", metadata={"foo": "bar"})
    assert chat_model.metadata == {"foo": "bar"}
    chat_model.complete(messages=[UserMessage("Say hello!")])
    # There are multiple callback calls due to streaming
    # Take the last one because the first is occasionally from another test
    callback_call = litellm_success_callback_calls[-1]
    assert callback_call["kwargs"]["litellm_params"]["metadata"] == {"foo": "bar"}


@pytest.mark.litellm_openai
def test_litellm_chat_model_custom_llm_provider(litellm_success_callback_calls):
    """Test that provided custom_llm_provider is passed to the litellm success callback."""
    chat_model = LitellmChatModel("gpt-4o", custom_llm_provider="openai")
    assert chat_model.custom_llm_provider == "openai"
    chat_model.complete(messages=[UserMessage("Say hello!")])
    callback_call = litellm_success_callback_calls[-1]
    assert callback_call["kwargs"]["litellm_params"]["custom_llm_provider"] == "openai"


@pytest.mark.litellm_openai
def test_litellm_chat_model_complete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = LitellmChatModel("gpt-4o")
    with pytest.raises(ToolSchemaParseError):
        chat_model.complete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


@pytest.fixture
def litellm_async_success_callback_calls() -> Iterator[list[dict[str, Any]]]:
    """A list of calls to the `async_log_success_event` callback"""
    original_success_callback = litellm.success_callback.copy()
    callback_calls: list[dict[str, Any]] = []

    class AddCallToList(CustomLogger):
        async def async_log_success_event(
            self, kwargs, response_obj, start_time, end_time
        ):
            callback_calls.append({"kwargs": kwargs})

    litellm.callbacks = [AddCallToList()]  # type: ignore[list-item]
    yield callback_calls
    litellm.callbacks = original_success_callback  # type: ignore[assignment]


@pytest.mark.litellm_openai
async def test_litellm_chat_model_metadata_async(litellm_async_success_callback_calls):
    """Test that provided metadata is passed to the litellm success callback."""
    chat_model = LitellmChatModel("gpt-4o", metadata={"foo": "bar"})
    assert chat_model.metadata == {"foo": "bar"}
    await chat_model.acomplete(messages=[UserMessage("Say hello!")])
    # Sleep to allow the callback to be called
    await asyncio.sleep(1)
    # There are multiple callback calls due to streaming
    # Take the last one because the first is occasionally from another test
    callback_call = litellm_async_success_callback_calls[-1]
    assert callback_call["kwargs"]["litellm_params"]["metadata"] == {"foo": "bar"}


@pytest.mark.litellm_openai
async def test_litellm_chat_model_custom_llm_provider_async(
    litellm_async_success_callback_calls,
):
    """Test that provided custom_llm_provider is passed to the litellm success callback."""
    chat_model = LitellmChatModel("gpt-4o", custom_llm_provider="openai")
    assert chat_model.custom_llm_provider == "openai"
    await chat_model.acomplete(messages=[UserMessage("Say hello!")])
    # Sleep to allow the callback to be called
    await asyncio.sleep(1)
    # There are multiple callback calls due to streaming
    # Take the last one because the first is occasionally from another test
    callback_call = litellm_async_success_callback_calls[-1]
    assert callback_call["kwargs"]["litellm_params"]["custom_llm_provider"] == "openai"


@pytest.mark.litellm_openai
async def test_litellm_chat_model_acomplete_raises_tool_schema_parse_error():
    def raise_error(v):
        raise ValueError(v)

    class Test(BaseModel):
        value: Annotated[int, AfterValidator(raise_error)]

    chat_model = LitellmChatModel("gpt-4o")
    with pytest.raises(ToolSchemaParseError):
        await chat_model.acomplete(
            messages=[UserMessage("Return a test value of 42.")],
            output_types=[Test],
        )


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        pytest.param(
            "Use the tool to return a list of three fruits",
            [list[str]],
            list,
            marks=pytest.mark.skip(reason="Claude fails to format list of strings."),
        ),
    ],
)
@pytest.mark.litellm_anthropic
def test_litellm_chat_model_complete_anthropic(
    prompt, output_types, expected_output_type
):
    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.litellm_anthropic
def test_litellm_chat_model_complete_anthropic_function_call():
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = chat_model.complete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.litellm_anthropic
def test_litellm_chat_model_complete_anthropic_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    def minus(a: int, b: int) -> int:
        return a - b

    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = chat_model.complete(
        messages=[
            UserMessage(
                "Use the plus tool to sum 1 and 2. Use the minus tool to subtract 1 from 2."
            )
        ],
        functions=[plus, minus],
        output_types=[ParallelFunctionCall[int]],
    )
    assert isinstance(message.content, ParallelFunctionCall)
    # Claude does not return multiple tool calls, so this returns a single function call


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        pytest.param(
            "Return True.",
            [bool],
            bool,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
        pytest.param(
            "Return [1, 2, 3, 4, 5]",
            [list[int]],
            list,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
        pytest.param(
            'Return ["apple", "banana"]',
            [list[str]],
            list,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
    ],
)
@pytest.mark.litellm_ollama
def test_litellm_chat_model_complete_ollama(prompt, output_types, expected_output_type):
    chat_model = LitellmChatModel(
        "ollama_chat/llama3.1", api_base="http://localhost:11434"
    )
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        ("List three fruits", [list[str]], list),
    ],
)
@pytest.mark.litellm_openai
async def test_litellm_chat_model_acomplete_openai(
    prompt, output_types, expected_output_type
):
    chat_model = LitellmChatModel("gpt-4o")
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        pytest.param(
            "Use the tool to return a list of three fruits",
            [list[str]],
            list,
            marks=pytest.mark.skip(reason="Claude fails to format list of strings."),
        ),
    ],
)
@pytest.mark.litellm_anthropic
async def test_litellm_chat_model_acomplete_anthropic(
    prompt, output_types, expected_output_type
):
    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.litellm_anthropic
async def test_litellm_chat_model_acomplete_anthropic_function_call():
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = await chat_model.acomplete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.litellm_anthropic
async def test_litellm_chat_model_acomplete_anthropic_async_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    def minus(a: int, b: int) -> int:
        return a - b

    chat_model = LitellmChatModel("anthropic/claude-3-haiku-20240307")
    message = await chat_model.acomplete(
        messages=[
            UserMessage(
                "Use the plus tool to sum 1 and 2. Use the minus tool to subtract 1 from 2."
            )
        ],
        functions=[plus, minus],
        output_types=[AsyncParallelFunctionCall[int]],
    )
    assert isinstance(message.content, AsyncParallelFunctionCall)
    # Claude does not return multiple tool calls, so this returns a single function call


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        pytest.param(
            "Return True.",
            [bool],
            bool,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
        pytest.param(
            "Return [1, 2, 3, 4, 5]",
            [list[int]],
            list,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
        pytest.param(
            'Return ["apple", "banana"]',
            [list[str]],
            list,
            marks=pytest.mark.skip(
                reason="Skip until Ollama supports streamed tool calls"
            ),
        ),
    ],
)
@pytest.mark.litellm_ollama
async def test_litellm_chat_model_acomplete_ollama(
    prompt, output_types, expected_output_type
):
    chat_model = LitellmChatModel(
        "ollama_chat/llama3.1", api_base="http://localhost:11434"
    )
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)



import typing
from collections.abc import Iterable
from types import NoneType
from typing import Any, Generic, TypeVar

import pytest
from pydantic import BaseModel

from magentic.typing import (
    is_any_origin_subclass,
    is_origin_abstract,
    is_origin_subclass,
    name_type,
    split_union_type,
)


@pytest.mark.parametrize(
    ("type_", "expected_types"),
    [
        (str, ["str"]),
        (int, ["int"]),
        (str | int, ["str", "int"]),
    ],
)
def test_split_union_type(type_, expected_types):
    assert [t.__name__ for t in split_union_type(type_)] == expected_types


@pytest.mark.parametrize(
    ("type_", "expected_result"),
    [
        (str, False),
        (list[str], False),
        (typing.Iterable[str], True),
    ],
)
def test_is_origin_abstract(type_, expected_result):
    assert is_origin_abstract(type_) == expected_result


@pytest.mark.parametrize(
    ("type_", "cls_or_tuple", "expected_result"),
    [
        (str, str, True),
        (str, int, False),
        (str, (str, int), True),
        (list[str], list, True),
        (list[str], Iterable, True),
        (list[str], int, False),
        (dict, dict, True),
        (dict[str, int], dict, True),
        (NoneType, dict, False),
        (Any, dict, False),
    ],
)
def test_is_origin_subclass(type_, cls_or_tuple, expected_result):
    assert is_origin_subclass(type_, cls_or_tuple) == expected_result


@pytest.mark.parametrize(
    ("types", "cls_or_tuple", "expected_result"),
    [
        ([str, int], str, True),
        ([list[str], str], list, True),
        ([list[str], str], int, False),
    ],
)
def test_is_any_origin_subclass(types, cls_or_tuple, expected_result):
    assert is_any_origin_subclass(types, cls_or_tuple) == expected_result


T1 = TypeVar("T1")
T2 = TypeVar("T2")


class GenericModel(BaseModel, Generic[T1]):
    one: T1


class DoubleGenericModel(BaseModel, Generic[T1, T2]):
    one: T1
    two: T2


@pytest.mark.parametrize(
    ("type_", "expected_name"),
    [
        (str, "str"),
        (int, "int"),
        (list[str], "list_of_str"),
        (str | bool, "str_or_bool"),
        (list[str | bool], "list_of_str_or_bool"),
        (list[str] | bool, "list_of_str_or_bool"),
        (dict[str, int], "dict_of_str_to_int"),
        (typing.Iterable[str], "iterable_of_str"),
        (GenericModel[int], "genericmodel_int"),
        (GenericModel[GenericModel[int]], "genericmodel_genericmodel_int"),
        (DoubleGenericModel[int, str], "doublegenericmodel_int_str"),
    ],
)
def test_name_type(type_, expected_name):
    assert name_type(type_) == expected_name



from typing import TYPE_CHECKING

import pytest

from magentic._chat import Chat
from magentic.chat_model.message import (
    AssistantMessage,
    FunctionResultMessage,
    UserMessage,
)
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import async_iter

if TYPE_CHECKING:
    from collections.abc import Awaitable


def test_chat_add_message():
    chat1 = Chat()
    chat2 = chat1.add_message(UserMessage(content="Hello"))
    assert chat1.messages == []
    assert chat2.messages == [UserMessage(content="Hello")]


def test_chat_last_message():
    chat = Chat([UserMessage(content="one"), UserMessage(content="two")])
    assert chat.last_message == UserMessage(content="two")


@pytest.mark.openai
def test_chat_submit():
    chat1 = Chat(
        messages=[UserMessage(content="Hello")],
    )
    chat2 = chat1.submit()
    assert chat1.messages == [UserMessage(content="Hello")]
    assert chat2.messages[0] == UserMessage(content="Hello")
    assert isinstance(chat2.messages[1], AssistantMessage)


@pytest.mark.openai
async def test_chat_asubmit():
    chat1 = Chat(
        messages=[UserMessage(content="Hello")],
    )
    chat2 = await chat1.asubmit()
    assert chat1.messages == [UserMessage(content="Hello")]
    assert chat2.messages[0] == UserMessage(content="Hello")
    assert isinstance(chat2.messages[1], AssistantMessage)


def test_exec_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    plus_1_2 = FunctionCall(plus, 1, 2)
    chat = Chat(
        messages=[
            UserMessage(content="What is 1 plus 2?"),
            AssistantMessage(content=plus_1_2),
        ],
        functions=[plus],
    )
    chat = chat.exec_function_call()
    assert len(chat.messages) == 3
    assert chat.messages[2] == FunctionResultMessage(3, plus_1_2)


def test_exec_function_call_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    plus_1_2 = FunctionCall(plus, 1, 2)
    plus_3_4 = FunctionCall(plus, 3, 4)
    chat = Chat(
        messages=[
            UserMessage(content="What is 1 plus 2? And 3 plus 4?"),
            AssistantMessage(content=ParallelFunctionCall([plus_1_2, plus_3_4])),
        ],
        functions=[plus],
    )
    chat = chat.exec_function_call()
    assert len(chat.messages) == 4
    assert chat.messages[2] == FunctionResultMessage(3, plus_1_2)
    assert chat.messages[3] == FunctionResultMessage(7, plus_3_4)


def test_exec_function_call_raises():
    def plus(a: int, b: int) -> int:
        return a + b

    chat = Chat(
        messages=[UserMessage(content="What is 1 plus 2?")],
        functions=[plus],
    )
    with pytest.raises(TypeError):
        chat = chat.exec_function_call()


async def test_aexec_function_call_async_function():
    async def aplus(a: int, b: int) -> int:
        return a + b

    aplus_1_2: FunctionCall[Awaitable[int]] = FunctionCall(aplus, 1, 2)
    chat = Chat(
        messages=[
            UserMessage(content="What is 1 plus 2?"),
            AssistantMessage(content=aplus_1_2),
        ],
        functions=[aplus],
    )
    chat = await chat.aexec_function_call()
    assert len(chat.messages) == 3
    assert chat.messages[2] == FunctionResultMessage(3, aplus_1_2)


async def test_aexec_function_call_not_async_function():
    def plus(a: int, b: int) -> int:
        return a + b

    plus_1_2 = FunctionCall(plus, 1, 2)
    chat = Chat(
        messages=[
            UserMessage(content="What is 1 plus 2?"),
            AssistantMessage(content=plus_1_2),
        ],
        functions=[plus],
    )
    chat = await chat.aexec_function_call()
    assert len(chat.messages) == 3
    assert chat.messages[2] == FunctionResultMessage(3, plus_1_2)


async def test_aexec_function_call_async_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    plus_1_2 = FunctionCall(plus, 1, 2)
    plus_3_4 = FunctionCall(plus, 3, 4)
    chat = Chat(
        messages=[
            UserMessage(content="What is 1 plus 2? And 3 plus 4?"),
            AssistantMessage(
                content=AsyncParallelFunctionCall(
                    async_iter([plus_1_2, plus_3_4])  # type: ignore[list-item]
                )
            ),
        ],
        functions=[plus],
    )
    chat = await chat.aexec_function_call()
    assert len(chat.messages) == 4
    assert chat.messages[2] == FunctionResultMessage(3, plus_1_2)
    assert chat.messages[3] == FunctionResultMessage(7, plus_3_4)


async def test_aexec_function_call_raises():
    async def aplus(a: int, b: int) -> int:
        return a + b

    chat = Chat(
        messages=[UserMessage(content="What is 1 plus 2?")],
        functions=[aplus],
    )
    with pytest.raises(TypeError):
        chat = await chat.aexec_function_call()



import json
import os
from collections.abc import Iterator
from itertools import count
from pathlib import Path
from typing import Any
from uuid import UUID

import pytest
from dotenv import load_dotenv
from pytest_mock import MockerFixture
from vcr import VCR
from vcr.request import Request


@pytest.fixture(autouse=True, scope="session")
def _load_dotenv():
    """Load .env file so API keys are available when rewriting VCR cassettes"""
    # Prefer env vars from .env over existing env vars
    load_dotenv(override=True)
    # Set default values so tests can run using existing cassettes without env vars
    load_dotenv(".env.template")
    # Remove URL from error messages to avoid them changing with pydantic version
    os.environ["PYDANTIC_ERRORS_INCLUDE_URL"] = "false"


def pytest_recording_configure(config: pytest.Config, vcr: VCR) -> None:
    """Register VCR matcher for JSON request bodies"""

    def is_json_body_equal(r1: Request, r2: Request) -> None:
        try:
            msg = "JSON body does not match"
            assert json.loads(r1.body) == json.loads(r2.body), msg  # type: ignore[arg-type, unused-ignore]
        except (TypeError, json.JSONDecodeError):
            return

    vcr.register_matcher("is_json_body_equal", is_json_body_equal)


@pytest.fixture(autouse=True, scope="session")
def vcr_config():
    """Configure VCR to match on JSON bodies and filter sensitive headers"""

    def before_record_response(response: dict[str, Any]) -> dict[str, Any]:
        filter_response_headers = [
            "openai-organization",
            "Set-Cookie",
        ]
        for response_header in filter_response_headers:
            response["headers"].pop(response_header, None)
        return response

    return {
        "match_on": ["method", "host", "path", "query", "is_json_body_equal"],
        "filter_headers": [
            # openai
            "authorization",
            "openai-organization",
            # anthropic
            "x-api-key",
            # litellm_openai
            "cookie",
        ],
        "before_record_response": before_record_response,
    }


def pytest_collection_modifyitems(
    session: pytest.Session, config: pytest.Config, items: list[pytest.Item]
) -> None:
    llm_markers = [
        "anthropic",
        "litellm_anthropic",
        "litellm_ollama",
        "litellm_openai",
        "mistral",
        "openai",
        "openai_ollama",
    ]
    for item in items:
        # Apply vcr marker to all LLM tests
        if any(marker in item.keywords for marker in llm_markers):
            item.add_marker(pytest.mark.vcr)


@pytest.fixture(autouse=True)
def _mock_create_unique_id(mocker: MockerFixture) -> None:
    """Mock `uuid4` to make `_create_unique_id` return deterministic values"""

    def _mock_uuid4() -> Iterator[UUID]:
        for i in count():
            yield UUID(int=i)

    mocker.patch("magentic.function_call.uuid4", side_effect=_mock_uuid4())


def test_mock_create_unique_id():
    """Test that `_mock_create_unique_id` makes `_create_unique_id` deterministic"""
    from magentic.function_call import _create_unique_id

    assert _create_unique_id() == "000000000"
    assert _create_unique_id() == "000000001"


@pytest.fixture
def document_bytes_pdf() -> bytes:
    return Path("tests/data/test.pdf").read_bytes()


@pytest.fixture
def image_bytes_jpg() -> bytes:
    return Path("tests/data/python-powered.jpg").read_bytes()


@pytest.fixture
def image_bytes_png() -> bytes:
    return Path("tests/data/python-powered.png").read_bytes()



import json
from collections.abc import AsyncIterator, Callable, Iterable, Iterator, Sequence
from enum import Enum
from functools import singledispatch
from itertools import groupby
from typing import Any, Generic, TypeVar, cast, overload

from magentic._parsing import contains_parallel_function_call_type, contains_string_type
from magentic._streamed_response import AsyncStreamedResponse, StreamedResponse
from magentic.chat_model.base import ChatModel, aparse_stream, parse_stream
from magentic.chat_model.function_schema import (
    BaseFunctionSchema,
    FunctionCallFunctionSchema,
    function_schema_for_type,
    get_async_function_schemas,
    get_function_schemas,
)
from magentic.chat_model.message import (
    AssistantMessage,
    DocumentBytes,
    ImageBytes,
    Message,
    SystemMessage,
    ToolResultMessage,
    Usage,
    UserMessage,
    _RawMessage,
)
from magentic.chat_model.stream import (
    AsyncOutputStream,
    FunctionCallChunk,
    OutputStream,
    StreamParser,
    StreamState,
)
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
    _create_unique_id,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr
from magentic.vision import UserImageMessage

try:
    import anthropic
    from anthropic.lib.streaming import MessageStreamEvent
    from anthropic.lib.streaming._messages import accumulate_event
    from anthropic.types import (
        DocumentBlockParam,
        ImageBlockParam,
        MessageParam,
        TextBlockParam,
        ToolChoiceParam,
        ToolChoiceToolParam,
        ToolParam,
        ToolUseBlockParam,
    )
except ImportError as error:
    msg = "To use AnthropicChatModel you must install the `anthropic` package using `pip install 'magentic[anthropic]'`."
    raise ImportError(msg) from error


class AnthropicMessageRole(Enum):
    ASSISTANT = "assistant"
    USER = "user"


@singledispatch
def message_to_anthropic_message(message: Message[Any]) -> MessageParam:
    """Convert a Message to an OpenAI message."""
    # TODO: Add instructions for registering new Message type to this error message
    raise NotImplementedError(type(message))


@singledispatch
async def async_message_to_anthropic_message(message: Message[Any]) -> MessageParam:
    """Async version of `message_to_anthropic_message`."""
    return message_to_anthropic_message(message)


@message_to_anthropic_message.register(_RawMessage)
def _(message: _RawMessage[Any]) -> MessageParam:
    # TODO: Validate the message content
    return message.content  # type: ignore[no-any-return]


@message_to_anthropic_message.register(UserMessage)
def _(message: UserMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        return {"role": AnthropicMessageRole.USER.value, "content": message.content}
    if isinstance(message.content, Iterable):
        content: list[TextBlockParam | DocumentBlockParam | ImageBlockParam] = []
        for block in message.content:
            if isinstance(block, str):
                content.append({"type": "text", "text": block})
            elif isinstance(block, DocumentBytes):
                content.append(
                    {
                        "type": "document",
                        "source": {
                            "type": "base64",
                            "media_type": block.mime_type,
                            "data": block.as_base64(),
                        },
                    }
                )
            elif isinstance(block, ImageBytes):
                content.append(
                    {
                        "type": "image",
                        "source": {
                            "type": "base64",
                            "media_type": block.mime_type,
                            "data": block.as_base64(),
                        },
                    }
                )
            else:
                msg = f"Invalid content type for UserMessage: {type(block)}"
                raise TypeError(msg)
        return {"role": AnthropicMessageRole.USER.value, "content": content}
    msg = f"Invalid content type for UserMessage: {type(message.content)}"
    raise TypeError(msg)


@message_to_anthropic_message.register(UserImageMessage)
def _(message: UserImageMessage[Any]) -> MessageParam:
    if not isinstance(message.content, bytes):
        msg = f"Invalid content type: {type(message.content)}"
        raise TypeError(msg)

    image_bytes = ImageBytes(message.content)
    return {
        "role": AnthropicMessageRole.USER.value,
        "content": [
            {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image_bytes.mime_type,
                    "data": image_bytes.as_base64(),
                },
            }
        ],
    }


def _function_call_to_tool_call_block(
    function_call: FunctionCall[Any],
) -> ToolUseBlockParam:
    function_schema = FunctionCallFunctionSchema(function_call.function)
    return {
        "type": "tool_use",
        "id": function_call._unique_id,
        "name": function_schema.name,
        "input": json.loads(function_schema.serialize_args(function_call)),
    }


@message_to_anthropic_message.register(AssistantMessage)
def _(message: AssistantMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": message.content,
        }

    if isinstance(message.content, FunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [_function_call_to_tool_call_block(message.content)],
        }

    if isinstance(message.content, ParallelFunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [
                _function_call_to_tool_call_block(function_call)
                for function_call in message.content
            ],
        }

    if isinstance(message.content, StreamedResponse):
        content_blocks: list[TextBlockParam | ToolUseBlockParam] = []
        for item in message.content:
            if isinstance(item, StreamedStr):
                content_blocks.append({"type": "text", "text": item.to_string()})
            elif isinstance(item, FunctionCall):
                content_blocks.append(_function_call_to_tool_call_block(item))
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": content_blocks,
        }

    function_schema = function_schema_for_type(type(message.content))
    return {
        "role": AnthropicMessageRole.ASSISTANT.value,
        "content": [
            {
                "type": "tool_use",
                # Can be random because no result will be inserted back into the chat
                "id": _create_unique_id(),
                "name": function_schema.name,
                "input": json.loads(function_schema.serialize_args(message.content)),
            }
        ],
    }


@async_message_to_anthropic_message.register(AssistantMessage)
async def _(message: AssistantMessage[Any]) -> MessageParam:
    if isinstance(message.content, AsyncParallelFunctionCall):
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": [
                _function_call_to_tool_call_block(function_call)
                async for function_call in message.content
            ],
        }

    if isinstance(message.content, AsyncStreamedResponse):
        content_blocks: list[TextBlockParam | ToolUseBlockParam] = []
        async for item in message.content:
            if isinstance(item, AsyncStreamedStr):
                content_blocks.append({"type": "text", "text": await item.to_string()})
            elif isinstance(item, FunctionCall):
                content_blocks.append(_function_call_to_tool_call_block(item))
        return {
            "role": AnthropicMessageRole.ASSISTANT.value,
            "content": content_blocks,
        }
    return message_to_anthropic_message(message)


@message_to_anthropic_message.register(ToolResultMessage)
def _(message: ToolResultMessage[Any]) -> MessageParam:
    if isinstance(message.content, str):
        content = message.content
    else:
        function_schema = function_schema_for_type(type(message.content))
        content = json.loads(function_schema.serialize_args(message.content))
    return {
        "role": AnthropicMessageRole.USER.value,
        "content": [
            {
                "type": "tool_result",
                "tool_use_id": message.tool_call_id,
                "content": content,
            }
        ],
    }


# TODO: Move this to the magentic level by allowing `UserMessage` have a list of content
def _combine_messages(messages: Iterable[MessageParam]) -> list[MessageParam]:
    """Combine messages with the same role, to get alternating roles.

    Alternating roles is a requirement of the Anthropic API.
    """
    combined_messages: list[MessageParam] = []
    for message_group in groupby(messages, lambda x: x["role"]):
        role, messages = message_group
        content = []
        for message in messages:
            if isinstance(message["content"], list):
                content.extend(message["content"])
            elif isinstance(message["content"], str):
                content.append({"type": "text", "text": message["content"]})
            else:
                content.append(message["content"])
        combined_messages.append({"role": role, "content": content})
    return combined_messages


T = TypeVar("T")
BaseFunctionSchemaT = TypeVar("BaseFunctionSchemaT", bound=BaseFunctionSchema[Any])


class BaseFunctionToolSchema(Generic[BaseFunctionSchemaT]):
    def __init__(self, function_schema: BaseFunctionSchemaT):
        self._function_schema = function_schema

    def to_dict(self) -> ToolParam:
        return {
            "name": self._function_schema.name,
            "description": self._function_schema.description or "",
            "input_schema": self._function_schema.parameters,
        }

    def as_tool_choice(self, *, disable_parallel_tool_use: bool) -> ToolChoiceToolParam:
        return {"type": "tool", "name": self._function_schema.name}


class AnthropicStreamParser(StreamParser[MessageStreamEvent]):
    def is_content(self, item: MessageStreamEvent) -> bool:
        return item.type == "content_block_delta"

    def get_content(self, item: MessageStreamEvent) -> str | None:
        if item.type == "text":
            return item.text
        return None

    def is_tool_call(self, item: MessageStreamEvent) -> bool:
        return (
            item.type == "content_block_start" and item.content_block.type == "tool_use"
        )

    def iter_tool_calls(self, item: MessageStreamEvent) -> Iterable[FunctionCallChunk]:
        if item.type == "content_block_start" and item.content_block.type == "tool_use":
            return [
                FunctionCallChunk(
                    id=item.content_block.id, name=item.content_block.name, args=None
                )
            ]
        if item.type == "input_json":
            return [FunctionCallChunk(id=None, name=None, args=item.partial_json)]
        return []


class AnthropicStreamState(StreamState[MessageStreamEvent]):
    def __init__(self) -> None:
        self._current_message_snapshot: anthropic.types.Message | None = (
            None  # TODO: type
        )
        self.usage_ref: list[Usage] = []

    def update(self, item: MessageStreamEvent) -> None:
        self._current_message_snapshot = accumulate_event(
            # Unrecognized event types are ignored
            event=item,  # type: ignore[arg-type]
            current_snapshot=self._current_message_snapshot,
        )
        if item.type == "message_stop":
            assert not self.usage_ref
            self.usage_ref.append(
                Usage(
                    input_tokens=item.message.usage.input_tokens,
                    output_tokens=item.message.usage.output_tokens,
                )
            )

    @property
    def current_message_snapshot(self) -> Message[Any]:
        assert self._current_message_snapshot is not None
        # TODO: Possible to return AssistantMessage here?
        return _RawMessage(self._current_message_snapshot.model_dump())


def _extract_system_message(
    messages: Iterable[Message[Any]],
) -> tuple[str | anthropic.NotGiven, list[Message[Any]]]:
    system_messages = [m for m in messages if isinstance(m, SystemMessage)]
    if len(system_messages) > 1:
        msg = "Only one system message is allowed per request."
        raise ValueError(msg)
    return (
        system_messages[0].content if system_messages else anthropic.NOT_GIVEN,
        [m for m in messages if not isinstance(m, SystemMessage)],
    )


def _if_given(value: T | None) -> T | anthropic.NotGiven:
    return value if value is not None else anthropic.NOT_GIVEN


R = TypeVar("R")


class AnthropicChatModel(ChatModel):
    """An LLM chat model that uses the `anthropic` python package."""

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        max_tokens: int = 1024,
        temperature: float | None = None,
    ):
        self._model = model
        self._api_key = api_key
        self._base_url = base_url
        self._max_tokens = max_tokens
        self._temperature = temperature

        self._client = anthropic.Anthropic(api_key=api_key, base_url=base_url)
        self._async_client = anthropic.AsyncAnthropic(
            api_key=api_key, base_url=base_url
        )

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_key(self) -> str | None:
        return self._api_key

    @property
    def base_url(self) -> str | None:
        return self._base_url

    @property
    def max_tokens(self) -> int:
        return self._max_tokens

    @property
    def temperature(self) -> float | None:
        return self._temperature

    @staticmethod
    def _get_tool_choice(
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type],
    ) -> ToolChoiceParam | anthropic.NotGiven:
        """Create the tool choice argument."""
        if contains_string_type(output_types):
            return anthropic.NOT_GIVEN
        disable_parallel_tool_use = not contains_parallel_function_call_type(
            output_types
        )
        if len(tool_schemas) == 1:
            return tool_schemas[0].as_tool_choice(
                disable_parallel_tool_use=disable_parallel_tool_use
            )
        return {"type": "any", "disable_parallel_tool_use": disable_parallel_tool_use}

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        if output_types is None:
            output_types = [] if functions else cast(list[type[R]], [str])

        function_schemas = get_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        system, messages = _extract_system_message(messages)

        response: Iterator[MessageStreamEvent] = self._client.messages.stream(
            model=self.model,
            messages=_combine_messages(
                [message_to_anthropic_message(m) for m in messages]
            ),
            max_tokens=self.max_tokens,
            stop_sequences=_if_given(stop),
            system=system,
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or anthropic.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
        ).__enter__()
        stream = OutputStream(
            response,
            function_schemas=function_schemas,
            parser=AnthropicStreamParser(),
            state=AnthropicStreamState(),
        )
        return AssistantMessage._with_usage(
            parse_stream(stream, output_types), usage_ref=stream.usage_ref
        )

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[R] | AssistantMessage[str]:
        """Async version of `complete`."""
        if output_types is None:
            output_types = [] if functions else cast(list[type[R]], [str])

        function_schemas = get_async_function_schemas(functions, output_types)
        tool_schemas = [BaseFunctionToolSchema(schema) for schema in function_schemas]

        system, messages = _extract_system_message(messages)

        response: AsyncIterator[
            MessageStreamEvent
        ] = await self._async_client.messages.stream(
            model=self.model,
            messages=_combine_messages(
                [await async_message_to_anthropic_message(m) for m in messages]
            ),
            max_tokens=self.max_tokens,
            stop_sequences=_if_given(stop),
            system=system,
            temperature=_if_given(self.temperature),
            tools=[schema.to_dict() for schema in tool_schemas] or anthropic.NOT_GIVEN,
            tool_choice=self._get_tool_choice(
                tool_schemas=tool_schemas, output_types=output_types
            ),
        ).__aenter__()
        stream = AsyncOutputStream(
            response,
            function_schemas=function_schemas,
            parser=AnthropicStreamParser(),
            state=AnthropicStreamState(),
        )
        return AssistantMessage._with_usage(
            await aparse_stream(stream, output_types), usage_ref=stream.usage_ref
        )



from typing import TYPE_CHECKING, Literal, cast
from unittest.mock import ANY, MagicMock

import pytest
from inline_snapshot import snapshot
from pydantic import BaseModel, TypeAdapter, ValidationError
from typing_extensions import assert_type

from magentic.chat_model.message import (
    AnyMessage,
    AssistantMessage,
    FunctionResultMessage,
    ImageBytes,
    ImageUrl,
    Placeholder,
    SystemMessage,
    ToolResultMessage,
    Usage,
    UserMessage,
)
from magentic.function_call import FunctionCall

if TYPE_CHECKING:
    from collections.abc import Sequence


def test_placeholder_format():
    class Country(BaseModel):
        name: str

    placeholder = Placeholder(Country, "country")

    assert_type(placeholder, Placeholder[Country])
    assert placeholder.name == "country"
    assert placeholder.format(country=Country(name="USA")) == Country(name="USA")


def test_placeholder_coercion():
    placeholder_str = Placeholder(str, "my_string")
    assert placeholder_str.format(my_string="test") == "test"
    placeholder_list_str = Placeholder(list[str], "my_list_str")
    assert placeholder_list_str.format(my_list_str=["test"]) == ["test"]
    assert placeholder_list_str.format(my_list_str={"test"}) == ["test"]  # set -> list
    with pytest.raises(ValueError):  # noqa: PT011
        placeholder_str.format(my_string=True)  # bool does not coerce to str


@pytest.mark.parametrize(
    ("message", "message_model_dump"),
    [
        (SystemMessage("Hello"), {"role": "system", "content": "Hello"}),
        (UserMessage("Hello"), {"role": "user", "content": "Hello"}),
        (AssistantMessage("Hello"), {"role": "assistant", "content": "Hello"}),
        (AssistantMessage(42), {"role": "assistant", "content": 42}),
        (
            ToolResultMessage(3, "unique_id"),
            {"role": "tool", "content": 3, "tool_call_id": "unique_id"},
        ),
        (
            FunctionResultMessage(3, FunctionCall(MagicMock(), 1, 2)),
            {"role": "tool", "content": 3, "tool_call_id": ANY},
        ),
    ],
)
def test_message_model_dump(message, message_model_dump):
    assert message.model_dump() == message_model_dump


@pytest.mark.parametrize(
    ("message", "message_repr"),
    [
        (SystemMessage("Hello"), "SystemMessage('Hello')"),
        (UserMessage("Hello"), "UserMessage('Hello')"),
        (AssistantMessage("Hello"), "AssistantMessage('Hello')"),
        (AssistantMessage(42), "AssistantMessage(42)"),
        (
            FunctionResultMessage(
                3, FunctionCall(MagicMock(__repr__=lambda x: "plus_repr"), 1, 2)
            ),
            "FunctionResultMessage(3, FunctionCall(plus_repr, 1, 2))",
        ),
    ],
)
def test_message_repr(message, message_repr):
    assert repr(message) == message_repr


def test_image_bytes_jpg(image_bytes_jpg: bytes) -> None:
    image_bytes = ImageBytes(image_bytes_jpg)
    assert image_bytes.mime_type == "image/jpeg"
    assert image_bytes.as_base64() == snapshot(
        "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"
    )


def test_image_bytes_png(image_bytes_png: bytes) -> None:
    image_bytes = ImageBytes(image_bytes_png)
    assert image_bytes.mime_type == "image/png"
    assert image_bytes.as_base64() == snapshot(
        " "
    )


def test_image_bytes_invalid():
    with pytest.raises(ValidationError):
        ImageBytes(b"invalid")


def test_user_message_format():
    user_message = UserMessage("Hello {x}")
    user_message_formatted = user_message.format(x="world")

    assert_type(user_message_formatted, UserMessage[str])
    assert_type(user_message_formatted.content, str)
    assert user_message_formatted == UserMessage("Hello world")


def test_user_message_format_type_hints():
    if TYPE_CHECKING:  # Avoid runtime error for None missing `format` method
        assert_type(cast(UserMessage[Literal["x"]], None).format(), UserMessage[str])
        # mypy does not convert `Literal` to `str` in these cases but pyright does
        # issue: https://github.com/python/mypy/issues/18419
        # assert_type(cast(UserMessage[Sequence[Literal["x"]]], None).format(), UserMessage[Sequence[str]])  # noqa: ERA001
        # assert_type(cast(UserMessage[Sequence[Literal["x"] | ImageBytes]], None).format(), UserMessage[Sequence[str | ImageBytes]])  # noqa: ERA001

        assert_type(cast(UserMessage[str], None).format(), UserMessage[str])

        assert_type(
            cast(UserMessage[Sequence[str]], None).format(), UserMessage[Sequence[str]]
        )
        assert_type(
            cast(UserMessage[str | Sequence[str]], None).format(),
            UserMessage[str | Sequence[str]],
        )

        assert_type(
            cast(UserMessage[Sequence[ImageBytes]], None).format(),
            UserMessage[Sequence[ImageBytes]],
        )
        assert_type(
            cast(UserMessage[Sequence[str | ImageBytes]], None).format(),
            UserMessage[Sequence[str | ImageBytes]],
        )
        assert_type(
            cast(UserMessage[str | Sequence[str | ImageBytes]], None).format(),
            UserMessage[str | Sequence[str | ImageBytes]],
        )

        assert_type(
            cast(UserMessage[Sequence[Placeholder[ImageUrl]]], None).format(),
            UserMessage[Sequence[ImageUrl]],
        )
        assert_type(
            cast(UserMessage[Sequence[str | Placeholder[ImageUrl]]], None).format(),
            UserMessage[Sequence[str | ImageUrl]],
        )
        assert_type(
            cast(
                UserMessage[Sequence[ImageBytes | Placeholder[ImageUrl]]], None
            ).format(),
            UserMessage[Sequence[ImageBytes | ImageUrl]],
        )
        assert_type(
            cast(
                UserMessage[Sequence[str | ImageBytes | Placeholder[ImageUrl]]], None
            ).format(),
            UserMessage[Sequence[str | ImageBytes | ImageUrl]],
        )

        assert_type(
            cast(UserMessage[str | Sequence[Placeholder[ImageUrl]]], None).format(),
            UserMessage[str | Sequence[ImageUrl]],
        )
        assert_type(
            cast(
                UserMessage[str | Sequence[str | Placeholder[ImageUrl]]], None
            ).format(),
            UserMessage[str | Sequence[str | ImageUrl]],
        )
        assert_type(
            cast(
                UserMessage[str | Sequence[ImageBytes | Placeholder[ImageUrl]]], None
            ).format(),
            UserMessage[str | Sequence[ImageBytes | ImageUrl]],
        )
        assert_type(
            cast(
                UserMessage[str | Sequence[str | ImageBytes | Placeholder[ImageUrl]]],
                None,
            ).format(),
            UserMessage[str | Sequence[str | ImageBytes | ImageUrl]],
        )


def test_assistant_message_usage():
    assistant_message = AssistantMessage("Hello")
    assert assistant_message.usage is None
    assistant_message = AssistantMessage._with_usage(
        "Hello", [Usage(input_tokens=1, output_tokens=2)]
    )
    assert assistant_message.usage == Usage(input_tokens=1, output_tokens=2)


def test_assistant_message_format_str():
    assistant_message = AssistantMessage("Hello {x}")
    assistant_message_formatted = assistant_message.format(x="world")

    assert_type(assistant_message_formatted, AssistantMessage[str])
    assert_type(assistant_message_formatted.content, str)
    assert assistant_message_formatted == AssistantMessage("Hello world")


def test_assistant_message_format_placeholder():
    class Country(BaseModel):
        name: str

    assistant_message = AssistantMessage(Placeholder(Country, "country"))
    assistant_message_formatted = assistant_message.format(country=Country(name="USA"))

    assert_type(assistant_message_formatted, AssistantMessage[Country])
    assert_type(assistant_message_formatted.content, Country)
    assert assistant_message_formatted == AssistantMessage(Country(name="USA"))


def test_assistant_message_format_type_hints():
    if TYPE_CHECKING:  # Avoid runtime error for None missing `format` method
        assert_type(cast(AssistantMessage[str], None).format(), AssistantMessage[str])
        assert_type(
            cast(AssistantMessage[Literal["x"]], None).format(), AssistantMessage[str]
        )
        assert_type(
            cast(AssistantMessage[Placeholder[int]], None).format(),
            AssistantMessage[int],
        )
        assert_type(
            cast(AssistantMessage[str | Placeholder[int]], None).format(),
            AssistantMessage[str | int],
        )
        assert_type(
            cast(AssistantMessage[Literal["x"] | Placeholder[int]], None).format(),
            AssistantMessage[str | int],
        )
        # Literal matches NotPlaceholder rather than str, so we cannot show that this will become str instead of Literal
        # assert_type(cast(AssistantMessage[Literal["x"] | int | Placeholder[int]], None).format(), AssistantMessage[str | int])  # noqa: ERA001


def test_function_result_message_eq():
    def plus(a: int, b: int) -> int:
        return a + b

    func_call = FunctionCall(plus, 1, 2)
    function_result_message = FunctionResultMessage(3, func_call)
    assert function_result_message == function_result_message
    assert function_result_message == FunctionResultMessage(3, func_call)
    # Different unique ids internally => not equal, despite equal FunctionCalls
    assert function_result_message != FunctionResultMessage(3, FunctionCall(plus, 1, 2))
    assert function_result_message != FunctionResultMessage(7, FunctionCall(plus, 3, 4))


def test_function_result_message_format():
    def plus(a: int, b: int) -> int:
        return a + b

    func_call = FunctionCall(plus, 1, 2)
    function_result_message = FunctionResultMessage(3, func_call)
    function_result_message_formatted = function_result_message.format(foo="bar")

    assert_type(function_result_message_formatted, FunctionResultMessage[int])
    assert_type(function_result_message_formatted.content, int)
    assert function_result_message_formatted == FunctionResultMessage(3, func_call)


def test_any_message():
    messages = [
        {"role": "system", "content": "Hello"},
        {"role": "user", "content": "Hello"},
        {"role": "assistant", "content": "Hello"},
        {"role": "tool", "content": 3, "tool_call_id": "unique_id"},
    ]
    assert TypeAdapter(list[AnyMessage]).validate_python(messages) == [
        SystemMessage("Hello"),
        UserMessage("Hello"),
        AssistantMessage("Hello"),
        ToolResultMessage(3, "unique_id"),
    ]



import inspect
from collections.abc import Callable, Sequence
from functools import wraps
from typing import Any, ParamSpec, TypeVar, cast

from magentic._chat import Chat
from magentic.chat_model.base import ChatModel
from magentic.chat_model.message import Message, UserMessage
from magentic.chatprompt import AsyncChatPromptFunction, ChatPromptFunction
from magentic.function_call import FunctionCall
from magentic.logger import logfire

P = ParamSpec("P")
R = TypeVar("R")


class MaxFunctionCallsError(Exception):
    """Raised when prompt chain reaches the max number of function calls."""


def prompt_chain(
    template: str | Sequence[Message[Any]],
    functions: list[Callable[..., Any]] | None = None,
    model: ChatModel | None = None,
    max_calls: int | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Convert a Python function to an LLM query, auto-resolving function calls.

    Use `@prompt_chain` when you need the LLM to perform multiple function calls to
    reach a final answer. When a function decorated with `@prompt_chain` is called, the
    LLM is queried, then any function calls are automatically executed and the results
    appended to the list of messages. Then the LLM is queried again and this repeats
    until a final answer is reached.

    Set `max_calls` to limit the number of function calls. If the limit is reached, a
    `MaxFunctionCallsError` will be raised.
    """

    messages = (
        [UserMessage(content=template)] if isinstance(template, str) else template
    )

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        func_signature = inspect.signature(func)

        if inspect.iscoroutinefunction(func):
            async_prompt_function = AsyncChatPromptFunction[P, Any](
                name=func.__name__,
                parameters=list(func_signature.parameters.values()),
                # TODO: Also allow ParallelFunctionCall. Support this more neatly
                return_type=func_signature.return_annotation | FunctionCall,  # type: ignore[arg-type,unused-ignore]
                messages=messages,
                functions=functions,
                model=model,
            )

            @wraps(func)
            async def awrapper(*args: P.args, **kwargs: P.kwargs) -> Any:
                with logfire.span(
                    f"Calling async prompt-chain {func.__name__}",
                    **func_signature.bind(*args, **kwargs).arguments,
                ):
                    chat = await Chat(
                        messages=async_prompt_function.format(*args, **kwargs),
                        functions=async_prompt_function.functions,
                        output_types=async_prompt_function.return_types,
                        model=async_prompt_function._model,  # Keep `None` value if unset
                    ).asubmit()
                    num_calls = 0
                    while isinstance(chat.last_message.content, FunctionCall):
                        if max_calls is not None and num_calls >= max_calls:
                            msg = (
                                f"Function {func.__name__} reached limit of"
                                f" {max_calls} function calls"
                            )
                            raise MaxFunctionCallsError(msg)
                        chat = await chat.aexec_function_call()
                        chat = await chat.asubmit()
                        num_calls += 1
                    return chat.last_message.content

            return cast(Callable[P, R], awrapper)

        prompt_function = ChatPromptFunction[P, R](
            name=func.__name__,
            parameters=list(func_signature.parameters.values()),
            # TODO: Also allow ParallelFunctionCall. Support this more neatly
            return_type=func_signature.return_annotation | FunctionCall,  # type: ignore[arg-type,unused-ignore]
            messages=messages,
            functions=functions,
            model=model,
        )

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            with logfire.span(
                f"Calling prompt-chain {func.__name__}",
                **func_signature.bind(*args, **kwargs).arguments,
            ):
                chat = Chat(
                    messages=prompt_function.format(*args, **kwargs),
                    functions=prompt_function.functions,
                    output_types=prompt_function.return_types,
                    model=prompt_function._model,  # Keep `None` value if unset
                ).submit()
                num_calls = 0
                while isinstance(chat.last_message.content, FunctionCall):
                    if max_calls is not None and num_calls >= max_calls:
                        msg = (
                            f"Function {func.__name__} reached limit of"
                            f" {max_calls} function calls"
                        )
                        raise MaxFunctionCallsError(msg)
                    chat = chat.exec_function_call().submit()
                    num_calls += 1
                return cast(R, chat.last_message.content)

        return wrapper

    return decorator



"""Tests for PromptFunction."""

from collections.abc import Awaitable
from inspect import getdoc
from typing import Annotated
from unittest.mock import AsyncMock, Mock

import pytest
from pydantic import AfterValidator, BaseModel

from magentic.chat_model.message import AssistantMessage, UserMessage
from magentic.chat_model.openai_chat_model import OpenaiChatModel
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.prompt_function import AsyncPromptFunction, PromptFunction, prompt
from magentic.settings import get_settings
from magentic.streaming import AsyncStreamedStr, StreamedStr


def test_promptfunction_format():
    @prompt("Test {param}.")
    def func(param: str) -> str: ...

    assert func.format("arg") == "Test arg."


def test_promptfunction_format_custom_type():
    class CustomType:
        def __format__(self, __format_spec: str) -> str:
            return "custom"

    @prompt("Test {param}.")
    def func(param: CustomType) -> str: ...

    assert func.format(CustomType()) == "Test custom."


def test_promptfunction_call():
    mock_model = Mock()
    mock_model.complete.return_value = AssistantMessage(content="Hello!")

    @prompt(
        "Hello {name}.",
        stop=["stop"],
        model=mock_model,
    )
    def say_hello(name: str) -> str | bool: ...

    assert say_hello("World") == "Hello!"
    assert mock_model.complete.call_count == 1
    assert mock_model.complete.call_args.kwargs["messages"] == [
        UserMessage("Hello World.")
    ]
    assert mock_model.complete.call_args.kwargs["output_types"] == [str, bool]
    assert mock_model.complete.call_args.kwargs["stop"] == ["stop"]


@pytest.mark.openai
def test_decorator_return_str():
    @prompt("What is the capital of {country}? Name only. No punctuation.")
    def get_capital(country: str) -> str:
        """This is the docstring."""

    assert isinstance(get_capital, PromptFunction)
    assert getdoc(get_capital) == "This is the docstring."
    output = get_capital("Ireland")
    assert isinstance(output, str)


@pytest.mark.openai
def test_decorator_return_bool():
    @prompt("True if {capital} is the capital of {country}.")
    def is_capital(capital: str, country: str) -> bool: ...

    assert is_capital("Dublin", "Ireland") is True


@pytest.mark.openai
def test_decorator_return_bool_str():
    @prompt("{text}")
    def query(text: str) -> bool | str: ...

    output = query("Hello, how are you?")
    assert isinstance(output, str)
    output = query("Use the tool/function to return the value True.")
    assert isinstance(output, bool)


@pytest.mark.skip(reason="Flaky")  # TODO: Make dict function call more reliable
@pytest.mark.openai
def test_decorator_return_dict():
    @prompt(
        "Ignore the defined inputs and pass inputs a=1, b=2 to the tool.",
        model=OpenaiChatModel("gpt-4"),
    )
    def return_mapping() -> dict[str, int]: ...

    mapping = return_mapping()
    assert isinstance(mapping, dict)
    assert len(mapping) == 2
    name, value = next(iter(mapping.items()))
    assert isinstance(name, str)
    assert isinstance(value, int)


@pytest.mark.openai
def test_decorator_return_pydantic_model():
    class CapitalCity(BaseModel):
        capital: str
        country: str

    @prompt("What is the capital of {country}? Make sure to follow the schema.")
    def get_capital(country: str) -> CapitalCity: ...

    output = get_capital("Ireland")
    assert isinstance(output, CapitalCity)


@pytest.mark.openai
def test_decorator_input_pydantic_model():
    class CapitalCity(BaseModel):
        capital: str
        country: str

    @prompt("Is this capital-country pair correct? {pair} Just answer True or False.")
    def check_capital(pair: CapitalCity) -> bool: ...

    assert check_capital(CapitalCity(capital="Dublin", country="Ireland"))


@pytest.mark.openai
def test_decorator_return_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    @prompt("Sum {a} and {b}", functions=[plus])
    def sum_ab(a: int, b: int) -> FunctionCall[int]: ...

    output = sum_ab(2, 3)
    assert isinstance(output, FunctionCall)
    func_result = output()
    assert isinstance(func_result, int)


@pytest.mark.openai
def test_decorator_return_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    def minus(a: int, b: int) -> int:
        return a - b

    @prompt("Sum {a} and {b}. Also subtract {a} from {b}.", functions=[plus, minus])
    def plus_and_minus(a: int, b: int) -> ParallelFunctionCall[int]: ...

    output = plus_and_minus(2, 3)
    assert isinstance(output, ParallelFunctionCall)
    func_result = output()
    assert len(func_result) == 2


@pytest.mark.openai
def test_decorator_ignore_multiple_tool_calls():
    """Test that when the model makes multiple tool calls, only the first is used."""

    # Provide two return types so that `tool_choice` does not force a single tool call
    @prompt(
        "Return the numbers 1 to 5 in the first tool call. And numbers 6 to 10 in the second.",
        model=OpenaiChatModel("gpt-4-1106-preview"),
    )
    def get_list() -> list[int] | bool: ...

    assert get_list() == [1, 2, 3, 4, 5]


@pytest.mark.openai
def test_decorator_return_streamed_str():
    @prompt("What is the capital of {country}?")
    def get_capital(country: str) -> StreamedStr: ...

    output = get_capital("Ireland")
    assert isinstance(output, StreamedStr)


@pytest.mark.openai
def test_decorator_max_retries():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    @prompt("Return a country.", max_retries=3)
    def get_country() -> Country: ...

    country = get_country()
    assert country.name == "Ireland"


async def test_async_promptfunction_call():
    mock_model = AsyncMock()
    mock_model.acomplete.return_value = AssistantMessage(content="Hello!")

    @prompt(
        "Hello {name}.",
        stop=["stop"],
        model=mock_model,
    )
    async def say_hello(name: str) -> str | bool: ...

    assert await say_hello("World") == "Hello!"
    assert mock_model.acomplete.call_count == 1
    assert mock_model.acomplete.call_args.kwargs["messages"] == [
        UserMessage("Hello World.")
    ]
    assert mock_model.acomplete.call_args.kwargs["output_types"] == [str, bool]
    assert mock_model.acomplete.call_args.kwargs["stop"] == ["stop"]


@pytest.mark.openai
async def test_async_decorator_return_str():
    @prompt("What is the capital of {country}? Name only. No punctuation.")
    async def get_capital(country: str) -> str: ...

    assert isinstance(get_capital, AsyncPromptFunction)
    output = await get_capital("Ireland")
    assert isinstance(output, str)


@pytest.mark.openai
async def test_async_decorator_return_async_streamed_str():
    @prompt("What is the capital of {country}?")
    async def get_capital(country: str) -> AsyncStreamedStr: ...

    output = await get_capital("Ireland")
    assert isinstance(output, AsyncStreamedStr)


@pytest.mark.openai
async def test_async_decorator_max_retries():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    @prompt("Return a country.", max_retries=3)
    async def get_country() -> Country: ...

    country = await get_country()
    assert country.name == "Ireland"


@pytest.mark.openai
async def test_async_decorator_return_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    @prompt("Sum {a} and {b}", functions=[plus])
    async def sum_ab(a: int, b: int) -> FunctionCall[int]: ...

    output = await sum_ab(2, 3)
    assert isinstance(output, FunctionCall)
    func_result = output()
    assert isinstance(func_result, int)


@pytest.mark.openai
async def test_async_decorator_return_async_function_call():
    async def async_plus(a: int, b: int) -> int:
        return a + b

    @prompt("Sum {a} and {b}", functions=[async_plus])
    async def sum_ab(a: int, b: int) -> FunctionCall[Awaitable[int]]: ...

    output = await sum_ab(2, 3)
    assert isinstance(output, FunctionCall)
    assert isinstance(await output(), int)


@pytest.mark.openai
async def test_decorator_return_async_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    async def minus(a: int, b: int) -> int:
        return a - b

    @prompt("Sum {a} and {b}. Also subtract {a} from {b}.", functions=[plus, minus])
    async def plus_and_minus(a: int, b: int) -> AsyncParallelFunctionCall[int]: ...

    output = await plus_and_minus(2, 3)
    assert isinstance(output, AsyncParallelFunctionCall)
    func_result = await output()
    assert len(func_result) == 2


def test_decorator_with_context_manager(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "test")

    @prompt("Say hello")
    def say_hello() -> str: ...

    @prompt(
        "Say hello",
        model=OpenaiChatModel("gpt-4", temperature=1),
    )
    def say_hello_gpt4() -> str: ...

    assert say_hello.model.model == get_settings().openai_model  # type: ignore[attr-defined]

    with OpenaiChatModel("gpt-4o"):
        assert say_hello.model.model == "gpt-4o"  # type: ignore[attr-defined]
        assert say_hello_gpt4.model.model == "gpt-4"  # type: ignore[attr-defined]



import base64
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Iterable, Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    NamedTuple,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    overload,
)

import filetype
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    RootModel,
    TypeAdapter,
    ValidationError,
    model_validator,
)
from typing_extensions import Self

from magentic.function_call import FunctionCall

if TYPE_CHECKING:
    from magentic.typing import NonStringSequence

PlaceholderTypeT = TypeVar("PlaceholderTypeT", covariant=True)


class Placeholder(BaseModel, Generic[PlaceholderTypeT]):
    """A placeholder for a value in a message.

    When formatting a message, the placeholder is replaced with the value.
    This is used in combination with the `@prompt`, `@promptchain`, and
    `@chatprompt` decorators to enable inserting function arguments into
    messages.
    """

    # TODO: Change to `type[PlaceholderT]` when pydantic supports it
    # issue: https://github.com/pydantic/pydantic/issues/9099
    type_: type
    name: str

    def __init__(self, type_: type[PlaceholderTypeT], name: str, **data: Any):
        super().__init__(type_=type_, name=name, **data)

    def format(self, **kwargs: Any) -> PlaceholderTypeT:
        if self.name not in kwargs:
            msg = f"Argument for {self.name!r} required by placeholder is missing"
            raise ValueError(msg)
        value = kwargs[self.name]
        type_adapter: TypeAdapter[PlaceholderTypeT] = TypeAdapter(self.type_)
        try:
            return type_adapter.validate_python(value)
        except ValidationError as e:
            msg = f"Argument for {self.name!r} must match placeholder type {self.type_!r} or be coercible to it"
            raise ValueError(msg) from e

    if TYPE_CHECKING:
        # HACK: Allows us to define protocol `NotPlaceholder`
        def __repr__(self) -> None: ...  # type: ignore[override]


class NotPlaceholder(Protocol):
    """Protocol that matches any type that is not a Placeholder."""

    # This matches all Python objects because they all have a `__repr__` method that
    # returns a string. However, to the type checker `Placeholder` appears to have a
    # `__repr__` method that returns `None`, so the protocol does not match it.
    def __repr__(self) -> str: ...


ContentT = TypeVar("ContentT", covariant=True)


class Message(BaseModel, Generic[ContentT], ABC):
    """A message sent to or from an LLM chat model."""

    content: ContentT

    def __init__(self, content: ContentT, **data: Any):
        super().__init__(content=content, **data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.content!r})"

    @abstractmethod
    def format(self, **kwargs: Any) -> "Message[Any]":
        """Format the message using the provided substitutions."""
        raise NotImplementedError


class _RawMessage(Message[ContentT], Generic[ContentT]):
    """A message that gets passed directly as a `message` object to the LLM provider.

    The content of this message should be a dict/object that matches the format
    expected by the LLM provider's Python client.
    """

    def __init__(self, content: ContentT, **data: Any):
        super().__init__(content=content, **data)

    # TODO: Add Usage to _RawMessage

    def format(self, **kwargs: Any) -> "_RawMessage[ContentT]":
        del kwargs
        return _RawMessage(self.content)


class SystemMessage(Message[str]):
    """A message to the LLM to guide the whole chat."""

    role: Literal["system"] = "system"

    def __init__(self, content: str, **data: Any):
        super().__init__(content=content, **data)

    def format(self, **kwargs: Any) -> "SystemMessage":
        return SystemMessage(self.content.format(**kwargs))


# Anthropic supports PDF: https://docs.anthropic.com/en/docs/build-with-claude/pdf-support
DocumentMimeType = Literal["application/pdf"]
_DOCUMENT_MIME_TYPES: tuple[DocumentMimeType, ...] = get_args(DocumentMimeType)


class DocumentBytes(RootModel[bytes]):
    """Bytes representing a document file."""

    @cached_property
    def mime_type(self) -> DocumentMimeType:
        mimetype: str | None = filetype.guess_mime(self.root)
        assert mimetype in _DOCUMENT_MIME_TYPES
        return cast(DocumentMimeType, mimetype)

    def __init__(self, root: bytes, **data: Any):
        super().__init__(root=root, **data)

    def as_base64(self) -> str:
        return base64.b64encode(self.root).decode("utf-8")

    def format(self, **kwargs: Any) -> Self:
        del kwargs
        return self

    @model_validator(mode="after")
    def _is_document_bytes(self) -> Self:
        mimetype: str | None = filetype.guess_mime(self.root)
        if mimetype not in _DOCUMENT_MIME_TYPES:
            msg = f"Unsupported document MIME type: {mimetype!r}"
            raise ValueError(msg)
        return self


# OpenAI supports PNG, JPEG, WEBP, and non-animated GIF
# Anthropic supports JPEG, PNG, GIF, or WebP
ImageMimeType = Literal["image/jpeg", "image/png", "image/gif", "image/webp"]
_IMAGE_MIME_TYPES: tuple[ImageMimeType, ...] = get_args(ImageMimeType)


class ImageBytes(RootModel[bytes]):
    """Bytes representing an image file."""

    @cached_property
    def mime_type(self) -> ImageMimeType:
        mimetype: str | None = filetype.guess_mime(self.root)
        assert mimetype in _IMAGE_MIME_TYPES
        return cast(ImageMimeType, mimetype)

    def as_base64(self) -> str:
        return base64.b64encode(self.root).decode("utf-8")

    def format(self, **kwargs: Any) -> Self:
        del kwargs
        return self

    @model_validator(mode="after")
    def _is_image_bytes(self) -> Self:
        mimetype: str | None = filetype.guess_mime(self.root)
        if mimetype not in _IMAGE_MIME_TYPES:
            msg = f"Unsupported image MIME type: {mimetype!r}"
            raise ValueError(msg)
        return self


class ImageUrl(RootModel[str]):
    """String representing a URL to an image."""

    def format(self, **kwargs: Any) -> Self:
        del kwargs
        return self


UserMessageContentBlock: TypeAlias = DocumentBytes | ImageBytes | ImageUrl
UserMessageContentT = TypeVar(
    "UserMessageContentT",
    bound=str
    | Sequence[str | UserMessageContentBlock | Placeholder[UserMessageContentBlock]],
    covariant=True,
)
UserMessageContentBlockT = TypeVar(
    "UserMessageContentBlockT", bound=UserMessageContentBlock
)
UserMessageContentBlockT2 = TypeVar(
    "UserMessageContentBlockT2", bound=UserMessageContentBlock
)
# These allow type hinting that a `str` _might_ be part of the union
StrT = TypeVar("StrT", str, str)
StrT2 = TypeVar("StrT2", str, str)


class UserMessage(Message[UserMessageContentT], Generic[UserMessageContentT]):
    """A message sent by a user to an LLM chat model."""

    role: Literal["user"] = "user"

    def __init__(self, content: UserMessageContentT, **data: Any):
        super().__init__(content=content, **data)

    @overload
    def format(self: "UserMessage[str]", **kwargs: Any) -> "UserMessage[str]": ...

    @overload
    def format(
        self: "UserMessage[StrT | NonStringSequence[StrT2 | UserMessageContentBlockT | Placeholder[UserMessageContentBlockT2]]]",
        **kwargs: Any,
    ) -> "UserMessage[StrT | Sequence[StrT2 | UserMessageContentBlockT | UserMessageContentBlockT2]]": ...

    def format(
        self: "UserMessage[StrT | NonStringSequence[StrT2 | UserMessageContentBlockT | Placeholder[UserMessageContentBlockT2]]]",
        **kwargs: Any,
    ) -> "UserMessage[StrT | Sequence[StrT2 | UserMessageContentBlockT | UserMessageContentBlockT2]]":
        if isinstance(self.content, str):
            return UserMessage(self.content.format(**kwargs))
        if isinstance(self.content, Iterable):
            return UserMessage([block.format(**kwargs) for block in self.content])  # type: ignore[misc]
        msg = f"Unsupported content type: {type(self.content)}"
        raise ValueError(msg)


class Usage(NamedTuple):
    """Usage statistics for the LLM request."""

    input_tokens: int
    output_tokens: int


T = TypeVar("T")
NotPlaceholderT = TypeVar("NotPlaceholderT", bound=NotPlaceholder)


class AssistantMessage(Message[ContentT], Generic[ContentT]):
    """A message received from an LLM chat model."""

    role: Literal["assistant"] = "assistant"
    _usage_ref: list[Usage] | None = PrivateAttr(None)

    def __init__(self, content: ContentT, **data: Any):
        super().__init__(content=content, **data)

    @classmethod
    def _with_usage(cls, content: ContentT, usage_ref: list[Usage]) -> Self:  # type: ignore[misc]
        message = cls(content)
        message._usage_ref = usage_ref
        return message

    @property
    def usage(self) -> Usage | None:
        if self._usage_ref:
            return self._usage_ref[0]
        return None

    @overload
    def format(
        self: "AssistantMessage[str]", **kwargs: Any
    ) -> "AssistantMessage[str]": ...

    # TODO: Combine with overload below (using StrT) when mypy respects typevar contraints
    # issue: https://github.com/python/mypy/issues/18419
    @overload
    def format(
        self: "AssistantMessage[Placeholder[T]]", **kwargs: Any
    ) -> "AssistantMessage[T]": ...

    @overload
    def format(
        self: "AssistantMessage[str | Placeholder[T]]", **kwargs: Any
    ) -> "AssistantMessage[str | T]": ...

    @overload
    def format(
        self: "AssistantMessage[NotPlaceholderT | Placeholder[T]]", **kwargs: Any
    ) -> "AssistantMessage[NotPlaceholderT | T]": ...

    def format(
        self: "AssistantMessage[str | NotPlaceholderT | Placeholder[T]]", **kwargs: Any
    ) -> "AssistantMessage[str | NotPlaceholderT | T]":
        if isinstance(self.content, str):
            return AssistantMessage(self.content.format(**kwargs))
        if isinstance(self.content, Placeholder):
            return AssistantMessage(self.content.format(**kwargs))
        return AssistantMessage(self.content)


class ToolResultMessage(Message[ContentT], Generic[ContentT]):
    """A message containing the result of a tool call."""

    role: Literal["tool"] = "tool"
    tool_call_id: str

    def __init__(self, content: ContentT, tool_call_id: str, **data: Any):
        super().__init__(content=content, tool_call_id=tool_call_id, **data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.content!r}, {self.tool_call_id=!r})"

    def format(self, **kwargs: Any) -> "ToolResultMessage[ContentT]":
        del kwargs
        return ToolResultMessage(self.content, self.tool_call_id)


class FunctionResultMessage(ToolResultMessage[ContentT], Generic[ContentT]):
    """A message containing the result of a function call."""

    _function_call: FunctionCall[Awaitable[ContentT]] | FunctionCall[ContentT]

    @overload
    def __init__(
        self,
        content: ContentT,
        function_call: FunctionCall[Awaitable[ContentT]],
        **data: Any,
    ): ...

    @overload
    def __init__(
        self, content: ContentT, function_call: FunctionCall[ContentT], **data: Any
    ): ...

    def __init__(
        self,
        content: ContentT,
        function_call: FunctionCall[Awaitable[ContentT]] | FunctionCall[ContentT],
        **data: Any,
    ):
        super().__init__(content=content, tool_call_id=function_call._unique_id, **data)
        self._function_call = function_call

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.content!r}, {self._function_call!r})"

    @property
    def function_call(
        self,
    ) -> FunctionCall[Awaitable[ContentT]] | FunctionCall[ContentT]:
        return self._function_call

    def format(self, **kwargs: Any) -> "FunctionResultMessage[ContentT]":
        del kwargs
        return FunctionResultMessage(self.content, self._function_call)


AnyMessage = Annotated[
    # Do not include FunctionResultMessage which also uses "tool" role
    SystemMessage | UserMessage[Any] | AssistantMessage[Any] | ToolResultMessage[Any],
    Field(discriminator="role"),
]
"""Union of all message types."""



from collections.abc import AsyncIterable, AsyncIterator, Iterable, Iterator
from typing import Any

from magentic.function_call import FunctionCall
from magentic.streaming import (
    AsyncStreamedStr,
    CachedAsyncIterable,
    CachedIterable,
    StreamedStr,
)


class StreamedResponse:
    """A streamed LLM response consisting of text output and tool calls.

    This is an iterable of StreamedStr and FunctionCall instances.

    Examples
    --------
    >>> from magentic import prompt, StreamedResponse, StreamedStr, FunctionCall
    >>>
    >>> def get_weather(city: str) -> str:
    >>>     return f"The weather in {city} is 20°C."
    >>>
    >>> @prompt(
    >>>     "Say hello, then get the weather for: {cities}",
    >>>     functions=[get_weather],
    >>> )
    >>> def describe_weather(cities: list[str]) -> StreamedResponse: ...
    >>>
    >>> response = describe_weather(["Cape Town", "San Francisco"])
    >>>
    >>> for item in response:
    >>>     if isinstance(item, StreamedStr):
    >>>         for chunk in item:
    >>>             print(chunk, sep="", end="")
    >>>         print()
    >>>     if isinstance(item, FunctionCall):
    >>>         print(item)
    >>>         print(item())
    Hello! I'll get the weather for Cape Town and San Francisco for you.
    FunctionCall(<function get_weather at 0x1109825c0>, 'Cape Town')
    The weather in Cape Town is 20°C.
    FunctionCall(<function get_weather at 0x1109825c0>, 'San Francisco')
    The weather in San Francisco is 20°C.
    """

    def __init__(self, stream: Iterable[StreamedStr | FunctionCall[Any]]):
        self._stream = CachedIterable(stream)

    def __iter__(self) -> Iterator[StreamedStr | FunctionCall[Any]]:
        yield from self._stream


class AsyncStreamedResponse:
    """Async version of `StreamedResponse`."""

    def __init__(self, stream: AsyncIterable[AsyncStreamedStr | FunctionCall[Any]]):
        self._stream = CachedAsyncIterable(stream)

    async def __aiter__(self) -> AsyncIterator[AsyncStreamedStr | FunctionCall[Any]]:
        async for item in self._stream:
            yield item



from typing import Annotated

import pytest
from pydantic import AfterValidator, BaseModel

from magentic.chat_model.anthropic_chat_model import AnthropicChatModel
from magentic.chat_model.litellm_chat_model import LitellmChatModel
from magentic.chat_model.message import UserMessage
from magentic.chat_model.openai_chat_model import OpenaiChatModel
from magentic.chat_model.retry_chat_model import RetryChatModel


@pytest.mark.openai
def test_retry_chat_model_complete_openai():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(OpenaiChatModel("gpt-4o-mini"), max_retries=3)
    message = chat_model.complete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"


@pytest.mark.openai
async def test_retry_chat_model_acomplete_openai():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(OpenaiChatModel("gpt-4o-mini"), max_retries=3)
    message = await chat_model.acomplete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"


@pytest.mark.anthropic
def test_retry_chat_model_complete_anthropic():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(
        AnthropicChatModel("claude-3-haiku-20240307"), max_retries=3
    )
    message = chat_model.complete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"


@pytest.mark.anthropic
async def test_retry_chat_model_acomplete_anthropic():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(
        AnthropicChatModel("claude-3-haiku-20240307"), max_retries=3
    )
    message = await chat_model.acomplete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"


@pytest.mark.litellm_openai
def test_retry_chat_model_complete_litellm_openai():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(LitellmChatModel("gpt-4o-mini"), max_retries=3)
    message = chat_model.complete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"


@pytest.mark.litellm_openai
async def test_retry_chat_model_acomplete_litellm_openai():
    def assert_is_ireland(v):
        if v != "Ireland":
            msg = "Country must be Ireland."
            raise ValueError(msg)
        return v

    class Country(BaseModel):
        name: Annotated[str, AfterValidator(assert_is_ireland)]

    chat_model = RetryChatModel(LitellmChatModel("gpt-4o-mini"), max_retries=3)
    message = await chat_model.acomplete(
        messages=[UserMessage("Return a country.")],
        output_types=[Country],
    )
    assert isinstance(message.content, Country)
    assert message.content.name == "Ireland"



from collections.abc import Callable, Iterable
from functools import singledispatchmethod
from typing import Any, TypeVar, overload

from magentic.chat_model.base import ChatModel, ToolSchemaParseError
from magentic.chat_model.message import AssistantMessage, Message, ToolResultMessage
from magentic.logger import logfire

R = TypeVar("R")


class RetryChatModel(ChatModel):
    """Wraps another ChatModel to add LLM-assisted retries."""

    def __init__(
        self,
        chat_model: ChatModel,
        *,
        max_retries: int,
    ):
        self._chat_model = chat_model
        self._max_retries = max_retries

    # TODO: Make this public to allow modifying error handling behavior
    # User should be able to add handlers to instance using decorator
    # e.g. `@my_retry_chat_model.exception_handler(exc_type)`
    # TODO: Add exception base class for those with output_message attribute
    @singledispatchmethod
    def _make_retry_messages(self, error: Exception) -> list[Message[Any]]:
        raise NotImplementedError

    # TODO: Catch UnknownToolError here
    @_make_retry_messages.register
    def _(self, error: ToolSchemaParseError) -> list[Message[Any]]:
        return [
            error.output_message,
            ToolResultMessage(
                content=str(error.validation_error), tool_call_id=error.tool_call_id
            ),
        ]

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        with logfire.span(
            "LLM-assisted retries enabled. Max {max_retries}",
            max_retries=self._max_retries,
        ):
            messages = list(messages)
            num_retry = 0
            while True:
                try:
                    message = self._chat_model.complete(
                        messages=messages,
                        functions=functions,
                        output_types=output_types,
                        stop=stop,
                    )
                # TODO: Get list of caught exceptions from _make_retry_messages registered types
                except ToolSchemaParseError as e:
                    if num_retry >= self._max_retries:
                        raise
                    messages += self._make_retry_messages(e)
                else:
                    return message

                num_retry += 1
                # TODO: Also log using Python logger
                logfire.warn(
                    "Retrying Chat Completion. Attempt {num_retry}",
                    num_retry=num_retry,
                )

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Async version of `complete`."""
        with logfire.span(
            "LLM-assisted retries enabled. Max {max_retries}",
            max_retries=self._max_retries,
        ):
            messages = list(messages)
            num_retry = 0
            while True:
                try:
                    message = await self._chat_model.acomplete(
                        messages=messages,
                        functions=functions,
                        output_types=output_types,
                        stop=stop,
                    )
                except ToolSchemaParseError as e:
                    if num_retry >= self._max_retries:
                        raise
                    messages += self._make_retry_messages(e)
                else:
                    return message

                num_retry += 1
                # TODO: Also log using Python logger
                logfire.warn(
                    "Retrying Chat Completion. Attempt {num_retry}",
                    num_retry=num_retry,
                )



from collections.abc import AsyncIterator

import pytest

from magentic import AsyncStreamedStr, StreamedStr
from magentic.streaming import (
    CachedAsyncIterable,
    CachedIterable,
    aapply,
    adropwhile,
    agroupby,
    aiter_streamed_json_array,
    apeek,
    apply,
    async_iter,
    atakewhile,
    azip,
    iter_streamed_json_array,
    peek,
)


async def test_async_iter():
    output = async_iter(["Hello", " World"])
    assert isinstance(output, AsyncIterator)
    assert [chunk async for chunk in output] == ["Hello", " World"]


def test_apply():
    items: list[int] = []
    iterable = apply(items.append, range(3))
    assert list(iterable) == [0, 1, 2]
    assert items == [0, 1, 2]


async def test_aapply():
    items: list[int] = []
    aiterable = aapply(items.append, async_iter(range(3)))
    assert [x async for x in aiterable] == [0, 1, 2]
    assert items == [0, 1, 2]


@pytest.mark.parametrize(
    ("aiterable", "expected"),
    [
        (azip(async_iter([1, 2, 3])), [(1,), (2,), (3,)]),
        (azip(async_iter([1, 2, 3]), async_iter([4, 5, 6])), [(1, 4), (2, 5), (3, 6)]),
    ],
)
async def test_azip(aiterable, expected):
    assert [x async for x in aiterable] == expected


@pytest.mark.parametrize(
    ("iterator", "expected_first", "expected_remaining"),
    [
        (iter([1, 2, 3]), 1, [1, 2, 3]),
        (iter([1]), 1, [1]),
    ],
)
def test_peek(iterator, expected_first, expected_remaining):
    first, remaining = peek(iterator)
    assert first == expected_first
    assert list(remaining) == expected_remaining


@pytest.mark.parametrize(
    ("aiterator", "expected_first", "expected_remaining"),
    [
        (async_iter([1, 2, 3]), 1, [1, 2, 3]),
        (async_iter([1]), 1, [1]),
    ],
)
async def test_apeek(aiterator, expected_first, expected_remaining):
    first, remaining = await apeek(aiterator)
    assert first == expected_first
    assert [x async for x in remaining] == expected_remaining


@pytest.mark.parametrize(
    ("predicate", "input", "expected"),
    [
        (lambda x: x < 3, async_iter(range(5)), [3, 4]),
        (lambda x: x < 6, async_iter(range(5)), []),
        (lambda x: x < 0, async_iter(range(5)), [0, 1, 2, 3, 4]),
    ],
)
async def test_adropwhile(predicate, input, expected):
    assert [x async for x in adropwhile(predicate, input)] == expected


@pytest.mark.parametrize(
    ("predicate", "input", "expected"),
    [
        (lambda x: x < 3, async_iter(range(5)), [0, 1, 2]),
        (lambda x: x < 6, async_iter(range(5)), [0, 1, 2, 3, 4]),
        (lambda x: x < 0, async_iter(range(5)), []),
    ],
)
async def test_atakewhile(predicate, input, expected):
    assert [x async for x in atakewhile(predicate, input)] == expected


@pytest.mark.parametrize(
    ("aiterable", "key", "expected"),
    [
        (async_iter([1, 1]), lambda x: x, [(1, [1, 1])]),
        (async_iter([1, 1, 2]), lambda x: x, [(1, [1, 1]), (2, [2])]),
    ],
)
async def test_agroupby(aiterable, key, expected):
    assert [
        (k, [x async for x in g]) async for k, g in agroupby(aiterable, key)
    ] == expected


async def test_agroupby_skip():
    aiterable = agroupby(async_iter([1, 1, 2, 2, 3, 3]), lambda x: x)
    key1, group1 = await anext(aiterable)
    key2, group2 = await anext(aiterable)
    assert [x async for x in group2] == [2, 2]


iter_streamed_json_array_test_cases = [
    (["[]"], []),
    (['["He', 'llo", ', '"Wo', 'rld"]'], ['"Hello"', '"World"']),
    (["[1, 2, 3]"], ["1", "2", "3"]),
    (["[1, ", "2, 3]"], ["1", "2", "3"]),
    (['[{"a": 1}, {2: "b"}]'], ['{"a": 1}', '{2: "b"}']),
    (["{\n", '"value', '":', " [", "1, ", "2, 3", "]"], ["1", "2", "3"]),
]


@pytest.mark.parametrize(("input", "expected"), iter_streamed_json_array_test_cases)
def test_iter_streamed_json_array(input, expected):
    assert list(iter_streamed_json_array(iter(input))) == expected


@pytest.mark.parametrize(("input", "expected"), iter_streamed_json_array_test_cases)
async def test_aiter_streamed_json_array(input, expected):
    assert [x async for x in aiter_streamed_json_array(async_iter(input))] == expected


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        ([1, 2, 3], [1, 2, 3]),
        (iter([1, 2, 3]), [1, 2, 3]),
        (range(3), [0, 1, 2]),
    ],
)
def test_iter_cached_iterable(input, expected):
    cached_iterable = CachedIterable(input)
    assert list(cached_iterable) == list(expected)
    assert list(cached_iterable) == list(expected)


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        ([1, 2, 3], [1, 2, 3]),
        (iter([1, 2, 3]), [1, 2, 3]),
        (range(3), [0, 1, 2]),
    ],
)
async def test_aiter_cached_async_iterable(input, expected):
    cached_aiterable = CachedAsyncIterable(async_iter(input))
    assert [x async for x in cached_aiterable] == list(expected)
    assert [x async for x in cached_aiterable] == list(expected)


def test_streamed_str_iter():
    iter_chunks = iter(["Hello", " World"])
    streamed_str = StreamedStr(iter_chunks)
    assert list(streamed_str) == ["Hello", " World"]
    assert list(iter_chunks) == []  # iterator is exhausted
    assert list(streamed_str) == ["Hello", " World"]


def test_streamed_str_str():
    streamed_str = StreamedStr(["Hello", " World"])
    assert str(streamed_str) == "Hello World"


def test_streamed_str_truncate():
    streamed_str = StreamedStr(["First", " Second", " Third"])
    assert streamed_str.truncate(length=12) == "First [...]"
    assert streamed_str.truncate(length=99) == "First Second Third"


async def test_async_streamed_str_iter():
    aiter_chunks = async_iter(["Hello", " World"])
    async_streamed_str = AsyncStreamedStr(aiter_chunks)
    assert [chunk async for chunk in async_streamed_str] == ["Hello", " World"]
    assert [chunk async for chunk in aiter_chunks] == []  # iterator is exhausted
    assert [chunk async for chunk in async_streamed_str] == ["Hello", " World"]


async def test_async_streamed_str_to_string():
    async_streamed_str = AsyncStreamedStr(async_iter(["Hello", " World"]))
    assert await async_streamed_str.to_string() == "Hello World"


async def test_async_streamed_str_truncate():
    async_streamed_str = AsyncStreamedStr(async_iter(["First", " Second", " Third"]))
    assert await async_streamed_str.truncate(length=12) == "First [...]"
    assert await async_streamed_str.truncate(length=99) == "First Second Third"



from unittest.mock import AsyncMock, Mock

import pytest

from magentic.chat_model.message import AssistantMessage, UserMessage
from magentic.function_call import FunctionCall
from magentic.prompt_chain import MaxFunctionCallsError, prompt_chain


@pytest.mark.openai
def test_prompt_chain():
    def get_current_weather(location, unit="fahrenheit"):
        """Get the current weather in a given location"""
        return {"temperature": "72", "forecast": ["sunny", "windy"]}

    @prompt_chain(
        template="What's the weather like in {city}?",
        functions=[get_current_weather],
    )
    def describe_weather(city: str) -> str: ...

    output = describe_weather("Boston")
    assert isinstance(output, str)


@pytest.mark.openai
def test_prompt_chain_messages():
    def get_current_weather(location, unit="fahrenheit"):
        """Get the current weather in a given location"""
        return {"temperature": "72", "forecast": ["sunny", "windy"]}

    @prompt_chain(
        template=[UserMessage("What's the weather like in {city}?")],
        functions=[get_current_weather],
    )
    def describe_weather(city: str) -> str: ...

    output = describe_weather("Boston")
    assert isinstance(output, str)


def test_prompt_chain_max_calls():
    mock_function = Mock()
    mock_function.__name__ = "mock_function_name"
    mock_model = Mock()
    mock_model.complete.return_value = AssistantMessage(
        content=FunctionCall(mock_function)
    )

    @prompt_chain(
        template="...",
        functions=[mock_function],
        model=mock_model,
        max_calls=1,
    )
    def make_function_call() -> str: ...

    with pytest.raises(MaxFunctionCallsError):
        make_function_call()
    assert mock_model.complete.call_count == 2
    assert mock_function.call_count == 1


@pytest.mark.openai
async def test_async_prompt_chain():
    async def get_current_weather(location, unit="fahrenheit"):
        """Get the current weather in a given location"""
        return {"temperature": "72", "forecast": ["sunny", "windy"]}

    @prompt_chain(
        template="What's the weather like in {city}?",
        functions=[get_current_weather],
    )
    async def describe_weather(city: str) -> str: ...

    output = await describe_weather("Boston")
    assert isinstance(output, str)


@pytest.mark.openai
async def test_async_prompt_chain_messages():
    async def get_current_weather(location, unit="fahrenheit"):
        """Get the current weather in a given location"""
        return {"temperature": "72", "forecast": ["sunny", "windy"]}

    @prompt_chain(
        template=[UserMessage("What's the weather like in {city}?")],
        functions=[get_current_weather],
    )
    async def describe_weather(city: str) -> str: ...

    output = await describe_weather("Boston")
    assert isinstance(output, str)


async def test_async_prompt_chain_max_calls():
    mock_function = Mock()
    mock_function.__name__ = "mock_function_name"
    mock_model = AsyncMock()
    mock_model.acomplete.return_value = AssistantMessage(
        content=FunctionCall(mock_function)
    )

    @prompt_chain(
        template="...",
        functions=[mock_function],
        model=mock_model,
        max_calls=1,
    )
    async def make_function_call() -> str: ...

    with pytest.raises(MaxFunctionCallsError):
        await make_function_call()
    assert mock_model.acomplete.call_count == 2
    assert mock_function.call_count == 1



"""Tests for @chatprompt decorator."""

from inspect import getdoc
from unittest.mock import AsyncMock, Mock

import pytest
from pydantic import BaseModel

from magentic.chat_model.message import (
    AssistantMessage,
    FunctionResultMessage,
    ImageBytes,
    Placeholder,
    SystemMessage,
    ToolResultMessage,
    UserMessage,
)
from magentic.chatprompt import (
    AsyncChatPromptFunction,
    ChatPromptFunction,
    chatprompt,
    escape_braces,
)
from magentic.function_call import FunctionCall


@pytest.mark.parametrize(
    "text",
    ["", "test", "{test}", "{{test}}", "{{test"],
)
def test_escape_braces(text):
    """Test that `escape_braces` makes `str.format` recover the original string."""
    assert escape_braces(text).format() == text


str_func_call = FunctionCall(lambda: "some string")


@pytest.mark.parametrize(
    ("message_templates", "expected_messages"),
    [
        (
            [AssistantMessage([1, 2, 3])],
            [AssistantMessage([1, 2, 3])],
        ),
        (
            [SystemMessage("System message with {param}.")],
            [SystemMessage("System message with arg.")],
        ),
        (
            [UserMessage("User message with {param}.")],
            [UserMessage("User message with arg.")],
        ),
        (
            [AssistantMessage("Assistant message with {param}.")],
            [AssistantMessage("Assistant message with arg.")],
        ),
        (
            [ToolResultMessage("Tool result message with {param}.", "unique_id")],
            [ToolResultMessage("Tool result message with {param}.", "unique_id")],
        ),
        # Do not format FunctionResultMessage
        (
            [
                FunctionResultMessage(
                    "Function result message with {param}", function_call=str_func_call
                )
            ],
            [
                FunctionResultMessage(
                    "Function result message with {param}", function_call=str_func_call
                )
            ],
        ),
    ],
)
def test_chatpromptfunction_format(message_templates, expected_messages):
    @chatprompt(*message_templates)
    def func(param: str) -> str: ...

    assert func.format(param="arg") == expected_messages


def test_chatpromptfunction_format_with_placeholder():
    class Country(BaseModel):
        name: str

    @chatprompt(
        AssistantMessage(Placeholder(Country, "country")),
    )
    def func(country: Country) -> str: ...

    assert func.format(Country(name="USA")) == [AssistantMessage(Country(name="USA"))]


def test_chatpromptfunction_call():
    mock_model = Mock()
    mock_model.complete.return_value = AssistantMessage(content="Hello!")

    @chatprompt(
        UserMessage("Hello {name}."),
        stop=["stop"],
        model=mock_model,
    )
    def say_hello(name: str) -> str | bool: ...

    assert say_hello("World") == "Hello!"
    assert mock_model.complete.call_count == 1
    assert mock_model.complete.call_args.kwargs["messages"] == [
        UserMessage("Hello World.")
    ]
    assert mock_model.complete.call_args.kwargs["output_types"] == [str, bool]
    assert mock_model.complete.call_args.kwargs["stop"] == ["stop"]


def test_chatprompt_decorator_docstring():
    @chatprompt(UserMessage("This is a user message."))
    def func(one: int) -> str:
        """This is the docstring."""
        ...

    assert isinstance(func, ChatPromptFunction)
    assert getdoc(func) == "This is the docstring."


async def test_asyncchatpromptfunction_call():
    mock_model = AsyncMock()
    mock_model.acomplete.return_value = AssistantMessage(content="Hello!")

    @chatprompt(
        UserMessage("Hello {name}."),
        stop=["stop"],
        model=mock_model,
    )
    async def say_hello(name: str) -> str | bool: ...

    assert await say_hello("World") == "Hello!"
    assert mock_model.acomplete.call_count == 1
    assert mock_model.acomplete.call_args.kwargs["messages"] == [
        UserMessage("Hello World.")
    ]
    assert mock_model.acomplete.call_args.kwargs["output_types"] == [str, bool]
    assert mock_model.acomplete.call_args.kwargs["stop"] == ["stop"]


async def test_async_chatprompt_decorator_docstring():
    @chatprompt(UserMessage("This is a user message."))
    async def func(one: int) -> str:
        """This is the docstring."""
        ...

    assert isinstance(func, AsyncChatPromptFunction)
    assert getdoc(func) == "This is the docstring."


@pytest.mark.openai
def test_chatprompt_readme_example():
    class Quote(BaseModel):
        quote: str
        character: str

    @chatprompt(
        SystemMessage("You are a movie buff."),
        UserMessage("What is your favorite quote from Harry Potter?"),
        AssistantMessage(
            Quote(
                quote="It does not do to dwell on dreams and forget to live.",
                character="Albus Dumbledore",
            )
        ),
        UserMessage("What is your favorite quote from {movie}?"),
    )
    def get_movie_quote(movie: str) -> Quote: ...

    movie_quote = get_movie_quote("Iron Man")
    assert isinstance(movie_quote, Quote)


@pytest.mark.openai
def test_chatprompt_with_function_call_and_result():
    def plus(a: int, b: int) -> int:
        return a + b

    plus_1_2 = FunctionCall(plus, 1, 2)

    @chatprompt(
        UserMessage("Use the plus function to add 1 and 2."),
        AssistantMessage(plus_1_2),
        FunctionResultMessage(3, plus_1_2),
    )
    def do_math() -> str: ...

    output = do_math()
    assert isinstance(output, str)


@pytest.mark.openai
def test_chatprompt_with_image_bytes(image_bytes_jpg):
    @chatprompt(
        UserMessage(["Describe this image in one word.", ImageBytes(image_bytes_jpg)]),
    )
    def describe_image() -> str: ...

    output = describe_image()
    assert isinstance(output, str)



import os
from collections.abc import Callable, Iterable, Sequence
from enum import Enum
from typing import Any, TypeVar, overload

import openai
from openai.types.chat import ChatCompletionStreamOptionsParam

from magentic._parsing import contains_string_type
from magentic.chat_model.base import ChatModel
from magentic.chat_model.message import AssistantMessage, Message
from magentic.chat_model.openai_chat_model import (
    BaseFunctionToolSchema,
    OpenaiChatModel,
)


class _MistralToolChoice(Enum):
    AUTO = "auto"
    """default mode. Model decides if it uses the tool or not."""
    ANY = "any"
    """forces tool use."""
    NONE = "none"
    """prevents tool use."""


class _MistralOpenaiChatModel(OpenaiChatModel):
    """Modified OpenaiChatModel to be compatible with Mistral API."""

    def _get_stream_options(self) -> ChatCompletionStreamOptionsParam | openai.NotGiven:
        return openai.NOT_GIVEN

    @staticmethod
    def _get_tool_choice(  # type: ignore[override]
        *,
        tool_schemas: Sequence[BaseFunctionToolSchema[Any]],
        output_types: Iterable[type],
    ) -> str | openai.NotGiven:
        """Create the tool choice argument.

        Mistral API has different options than the OpenAI API for `tool_choice`.
        See https://docs.mistral.ai/capabilities/function_calling/#tool_choice
        """
        if contains_string_type(output_types):
            return openai.NOT_GIVEN
        return _MistralToolChoice.ANY.value

    def _get_parallel_tool_calls(
        self, *, tools_specified: bool, output_types: Iterable[type]
    ) -> bool | openai.NotGiven:
        return openai.NOT_GIVEN


R = TypeVar("R")


class MistralChatModel(ChatModel):
    """An LLM chat model for the Mistral API.

    Currently this uses the `openai` Python package. This may change in the future.
    """

    def __init__(
        self,
        model: str,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        max_tokens: int | None = None,
        seed: int | None = None,
        temperature: float | None = None,
    ):
        self._mistral_openai_chat_model = _MistralOpenaiChatModel(
            model,
            api_key=api_key or os.getenv("MISTRAL_API_KEY"),
            base_url=base_url or "https://api.mistral.ai/v1/",
            max_tokens=max_tokens,
            seed=seed,
            temperature=temperature,
        )

    @property
    def model(self) -> str:
        return self._mistral_openai_chat_model.model

    @property
    def api_key(self) -> str | None:
        return self._mistral_openai_chat_model.api_key

    @property
    def base_url(self) -> str | None:
        return self._mistral_openai_chat_model.base_url

    @property
    def max_tokens(self) -> int | None:
        return self._mistral_openai_chat_model.max_tokens

    @property
    def seed(self) -> int | None:
        return self._mistral_openai_chat_model.seed

    @property
    def temperature(self) -> float | None:
        return self._mistral_openai_chat_model.temperature

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    def complete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[str] | AssistantMessage[R]:
        """Request an LLM message."""
        return self._mistral_openai_chat_model.complete(
            messages=messages,
            functions=functions,
            output_types=output_types,
            stop=stop,
        )

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: None = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[str]: ...

    @overload
    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Any = ...,
        output_types: Iterable[type[R]] = ...,
        *,
        stop: list[str] | None = ...,
    ) -> AssistantMessage[R]: ...

    async def acomplete(
        self,
        messages: Iterable[Message[Any]],
        functions: Iterable[Callable[..., Any]] | None = None,
        output_types: Iterable[type[R]] | None = None,
        *,
        stop: list[str] | None = None,
    ) -> AssistantMessage[R] | AssistantMessage[str]:
        """Async version of `complete`."""
        return await self._mistral_openai_chat_model.acomplete(
            messages=messages,
            functions=functions,
            output_types=output_types,
            stop=stop,
        )



import pytest
from pydantic import BaseModel

from magentic.chat_model.message import (
    AssistantMessage,
    SystemMessage,
    Usage,
    UserMessage,
)
from magentic.chat_model.mistral_chat_model import MistralChatModel
from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import AsyncStreamedStr, StreamedStr


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        ("List three fruits", [list[str]], list),
    ],
)
@pytest.mark.mistral
def test_mistral_chat_model_complete(prompt, output_types, expected_output_type):
    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.mistral
def test_mistral_chat_model_complete_usage():
    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[UserMessage("Say hello!")], output_types=[StreamedStr]
    )
    str(message.content)  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.mistral
def test_mistral_chat_model_complete_usage_structured_output():
    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[UserMessage("Count to 5")], output_types=[list[int]]
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.mistral
def test_mistral_chat_model_complete_function_call():
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.mistral
def test_mistral_chat_model_complete_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    def minus(a: int, b: int) -> int:
        return a - b

    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[
            UserMessage(
                "Use the plus tool to sum 1 and 2."
                " Use the minus tool to subtract 1 from 2."
                " Make sure to use both tools at once."
            )
        ],
        functions=[plus, minus],
        output_types=[ParallelFunctionCall[int]],
    )
    assert isinstance(message.content, ParallelFunctionCall)
    assert len(list(message.content)) == 2


@pytest.mark.mistral
def test_mistral_chat_model_few_shot_prompt():
    class Quote(BaseModel):
        quote: str
        character: str

    chat_model = MistralChatModel("mistral-large-latest")
    message = chat_model.complete(
        messages=[
            SystemMessage("You are a movie buff."),
            UserMessage("What is your favorite quote from Harry Potter?"),
            AssistantMessage(
                Quote(
                    quote="It does not do to dwell on dreams and forget to live.",
                    character="Albus Dumbledore",
                )
            ),
            # Mistral requires AssistantMessage after tool output
            # TODO: Automatically add this in ChatModel.complete like for tool calls
            AssistantMessage("."),
            UserMessage("What is your favorite quote from {movie}?"),
        ],
        output_types=[Quote],
    )
    assert isinstance(message.content, Quote)


@pytest.mark.parametrize(
    ("prompt", "output_types", "expected_output_type"),
    [
        ("Say hello!", [str], str),
        ("Return True", [bool], bool),
        ("Return the numbers 1 to 5", [list[int]], list),
        ("List three fruits", [list[str]], list),
    ],
)
@pytest.mark.mistral
async def test_mistral_chat_model_acomplete(prompt, output_types, expected_output_type):
    chat_model = MistralChatModel("mistral-large-latest")
    message = await chat_model.acomplete(
        messages=[UserMessage(prompt)], output_types=output_types
    )
    assert isinstance(message.content, expected_output_type)


@pytest.mark.mistral
async def test_mistral_chat_model_acomplete_usage():
    chat_model = MistralChatModel("mistral-large-latest")
    message = await chat_model.acomplete(
        messages=[UserMessage("Say hello!")], output_types=[AsyncStreamedStr]
    )
    await message.content.to_string()  # Finish the stream
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.mistral
async def test_mistral_chat_model_acomplete_usage_structured_output():
    chat_model = MistralChatModel("mistral-large-latest")
    message = await chat_model.acomplete(
        messages=[UserMessage("Count to 5")], output_types=[list[int]]
    )
    assert isinstance(message.usage, Usage)
    assert message.usage.input_tokens > 0
    assert message.usage.output_tokens > 0


@pytest.mark.mistral
async def test_mistral_chat_model_acomplete_function_call():
    def plus(a: int, b: int) -> int:
        """Sum two numbers."""
        return a + b

    chat_model = MistralChatModel("mistral-large-latest")
    message = await chat_model.acomplete(
        messages=[UserMessage("Use the tool to sum 1 and 2")],
        functions=[plus],
        output_types=[FunctionCall[int]],
    )
    assert isinstance(message.content, FunctionCall)


@pytest.mark.mistral
async def test_mistral_chat_model_acomplete_async_parallel_function_call():
    def plus(a: int, b: int) -> int:
        return a + b

    def minus(a: int, b: int) -> int:
        return a - b

    chat_model = MistralChatModel("mistral-large-latest")
    message = await chat_model.acomplete(
        messages=[
            UserMessage(
                "Use the plus tool to sum 1 and 2."
                " Use the minus tool to subtract 1 from 2."
                " Make sure to use both tools at once."
            )
        ],
        functions=[plus, minus],
        output_types=[AsyncParallelFunctionCall[int]],
    )
    assert isinstance(message.content, AsyncParallelFunctionCall)
    assert len([x async for x in message.content]) == 2



import copy
import inspect
from collections.abc import Awaitable, Callable, Sequence
from functools import update_wrapper
from typing import Any, Generic, ParamSpec, Protocol, TypeVar, cast, overload

from magentic.backend import get_chat_model
from magentic.chat_model.base import ChatModel
from magentic.chat_model.message import UserMessage
from magentic.chat_model.retry_chat_model import RetryChatModel
from magentic.logger import logfire
from magentic.typing import split_union_type

P = ParamSpec("P")
# TODO: Make `R` type Union of all possible return types except FunctionCall ?
# Then `R | FunctionCall[FuncR]` will separate FunctionCall from other return types.
# Can then use `FuncR` to make typechecker check `functions` argument to `prompt`
# `Not` type would solve this - https://github.com/python/typing/issues/801
R = TypeVar("R")


class BasePromptFunction(Generic[P, R]):
    """Base class for an LLM prompt template that is directly callable to query the LLM."""

    def __init__(
        self,
        name: str,
        parameters: Sequence[inspect.Parameter],
        return_type: type[R],
        template: str,
        functions: list[Callable[..., Any]] | None = None,
        stop: list[str] | None = None,
        max_retries: int = 0,
        model: ChatModel | None = None,
    ):
        self._name = name
        self._signature = inspect.Signature(
            parameters=parameters,
            return_annotation=return_type,
        )
        self._template = template
        self._functions = functions or []
        self._stop = stop
        self._max_retries = max_retries
        self._model = model

        self._return_types = list(split_union_type(return_type))

    @property
    def functions(self) -> list[Callable[..., Any]]:
        return self._functions.copy()

    @property
    def stop(self) -> list[str] | None:
        return copy.copy(self._stop)

    @property
    def model(self) -> ChatModel:
        if self._max_retries:
            return RetryChatModel(
                chat_model=self._model or get_chat_model(),
                max_retries=self._max_retries,
            )
        return self._model or get_chat_model()

    @property
    def return_types(self) -> list[type[R]]:
        return self._return_types.copy()

    def format(self, *args: P.args, **kwargs: P.kwargs) -> str:
        """Format the prompt template with the given arguments."""
        bound_args = self._signature.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return self._template.format(**bound_args.arguments)


class PromptFunction(BasePromptFunction[P, R], Generic[P, R]):
    """An LLM prompt template that is directly callable to query the LLM."""

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Query the LLM with the formatted prompt template."""
        with logfire.span(
            f"Calling prompt-function {self._name}",
            **self._signature.bind(*args, **kwargs).arguments,
        ):
            message = self.model.complete(
                messages=[UserMessage(content=self.format(*args, **kwargs))],
                functions=self._functions,
                output_types=self._return_types,
                stop=self._stop,
            )
            return message.content


class AsyncPromptFunction(BasePromptFunction[P, R], Generic[P, R]):
    """Async version of `PromptFunction`."""

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """Asynchronously query the LLM with the formatted prompt template."""
        with logfire.span(
            f"Calling async prompt-function {self._name}",
            **self._signature.bind(*args, **kwargs).arguments,
        ):
            message = await self.model.acomplete(
                messages=[UserMessage(content=self.format(*args, **kwargs))],
                functions=self._functions,
                output_types=self._return_types,
                stop=self._stop,
            )
            return message.content


class PromptDecorator(Protocol):
    """Protocol for a decorator that returns a `PromptFunction`.

    This allows finer-grain type annotation of the `prompt` function
    See https://github.com/microsoft/pyright/issues/5014#issuecomment-1523778421
    """

    @overload
    def __call__(self, func: Callable[P, Awaitable[R]]) -> AsyncPromptFunction[P, R]:  # type: ignore[overload-overlap]
        ...

    @overload
    def __call__(self, func: Callable[P, R]) -> PromptFunction[P, R]: ...


def prompt(
    template: str,
    functions: list[Callable[..., Any]] | None = None,
    stop: list[str] | None = None,
    max_retries: int = 0,
    model: ChatModel | None = None,
) -> PromptDecorator:
    """Convert a function into an LLM prompt template.

    The `@prompt` decorator allows you to define a template for a Large Language Model (LLM) prompt as a function.
    When this function is called, the arguments are inserted into the template, then this prompt is sent to an LLM which
    generates the function output.

    Examples
    --------
    >>> @prompt("Add more dudeness to: {phrase}")
    >>> def dudeify(phrase: str) -> str: ...
    >>> # No function body as this is never executed
    >>>
    >>> dudeify("Hello, how are you?")
    "Hey, dude! What's up? How's it going, my man?"
    """

    def decorator(
        func: Callable[P, Awaitable[R]] | Callable[P, R],
    ) -> AsyncPromptFunction[P, R] | PromptFunction[P, R]:
        func_signature = inspect.signature(func)

        if inspect.iscoroutinefunction(func):
            async_prompt_function = AsyncPromptFunction[P, R](
                name=func.__name__,
                parameters=list(func_signature.parameters.values()),
                return_type=func_signature.return_annotation,
                template=template,
                functions=functions,
                stop=stop,
                max_retries=max_retries,
                model=model,
            )
            return cast(
                AsyncPromptFunction[P, R],
                update_wrapper(async_prompt_function, func),
            )

        prompt_function = PromptFunction[P, R](
            name=func.__name__,
            parameters=list(func_signature.parameters.values()),
            return_type=func_signature.return_annotation,
            template=template,
            functions=functions,
            stop=stop,
            max_retries=max_retries,
            model=model,
        )
        return cast(PromptFunction[P, R], update_wrapper(prompt_function, func))

    return cast(PromptDecorator, decorator)



from abc import ABC, abstractmethod
from collections.abc import AsyncIterable, AsyncIterator, Iterable, Iterator
from itertools import chain
from typing import Any, Generic, NamedTuple, TypeVar

from pydantic import ValidationError

from magentic.chat_model.base import ToolSchemaParseError, UnknownToolError
from magentic.chat_model.function_schema import (
    AsyncFunctionSchema,
    FunctionSchema,
    select_function_schema,
)
from magentic.chat_model.message import Message, Usage
from magentic.streaming import (
    AsyncStreamedStr,
    StreamedStr,
    aapply,
    achain,
    aconsume,
    apply,
    async_iter,
    consume,
)

ItemT = TypeVar("ItemT")
OutputT = TypeVar("OutputT")


class FunctionCallChunk(NamedTuple):
    id: str | None
    name: str | None
    args: str | None


class StreamParser(ABC, Generic[ItemT]):
    @abstractmethod
    def is_content(self, item: ItemT) -> bool: ...

    @abstractmethod
    def get_content(self, item: ItemT) -> str | None: ...

    @abstractmethod
    def is_tool_call(self, item: ItemT) -> bool: ...

    @abstractmethod
    def iter_tool_calls(self, item: ItemT) -> Iterable[FunctionCallChunk]: ...


class StreamState(ABC, Generic[ItemT]):
    """Tracks the state of the LLM output stream.

    - message snapshot
    - usage
    - stop reason
    """

    usage_ref: list[Usage]

    @abstractmethod
    def update(self, item: ItemT) -> None: ...

    @property
    @abstractmethod
    def current_message_snapshot(self) -> Message[Any]: ...


class OutputStream(Generic[ItemT, OutputT]):
    """Converts streamed LLM output into a stream of magentic objects."""

    def __init__(
        self,
        stream: Iterator[ItemT],
        function_schemas: Iterable[FunctionSchema[OutputT]],
        parser: StreamParser[ItemT],
        state: StreamState[ItemT],
    ):
        self._stream = stream
        self._function_schemas = function_schemas
        self._parser = parser
        self._state = state

        self._iterator = self.__stream__()
        self._exhausted: bool = False

    def __next__(self) -> StreamedStr | OutputT:
        return self._iterator.__next__()

    def __iter__(self) -> Iterator[StreamedStr | OutputT]:
        yield from self._iterator

    def _streamed_str(
        self, stream: Iterator[ItemT], current_item_ref: list[ItemT]
    ) -> Iterator[str]:
        for item in stream:
            if content := self._parser.get_content(item):
                yield content
            if self._parser.is_tool_call(item):
                # TODO: Check if output types allow for early return and raise if not
                assert not current_item_ref
                current_item_ref.append(item)
                return
        self._exhausted = True

    def _tool_call(
        self,
        stream: Iterator[FunctionCallChunk],
        current_tool_call_ref: list[FunctionCallChunk],
        current_tool_call_id: str,
    ) -> Iterator[str]:
        for item in stream:
            # Only end the stream if we encounter a new tool call
            # so that the whole stream is consumed including stop_reason/usage chunks
            if item.id and item.id != current_tool_call_id:
                # TODO: Check if output types allow for early return and raise if not
                assert not current_tool_call_ref
                current_tool_call_ref.append(item)
                return
            if item.args:
                yield item.args
        self._exhausted = True

    def __stream__(self) -> Iterator[StreamedStr | OutputT]:
        # This works similarly to `itertools.groupby`
        stream = apply(self._state.update, self._stream)
        current_item_ref = [next(stream)]
        while current_item_ref:
            current_item = current_item_ref.pop()
            if self._parser.is_content(current_item):
                stream = chain([current_item], stream)
                streamed_str = StreamedStr(self._streamed_str(stream, current_item_ref))
                yield streamed_str
                if not current_item_ref and not self._exhausted:
                    # Finish the group to allow advancing to the next one
                    # Consume stream via StreamedStr so it can cache
                    consume(streamed_str)
            elif self._parser.is_tool_call(current_item):
                tool_calls_stream: Iterator[FunctionCallChunk] = (
                    tool_call_chunk
                    for item in chain([current_item], stream)
                    for tool_call_chunk in self._parser.iter_tool_calls(item)
                )
                tool_call_ref = [next(tool_calls_stream)]
                while tool_call_ref:
                    current_tool_call_chunk = tool_call_ref.pop()
                    current_tool_call_id = current_tool_call_chunk.id
                    assert current_tool_call_id is not None
                    assert current_tool_call_chunk.name is not None
                    function_schema = select_function_schema(
                        self._function_schemas, current_tool_call_chunk.name
                    )
                    if function_schema is None:
                        assert current_tool_call_id is not None
                        raise UnknownToolError(
                            output_message=self._state.current_message_snapshot,
                            tool_call_id=current_tool_call_id,
                            tool_name=current_tool_call_chunk.name,
                        )
                    try:
                        tool_calls_stream = chain(
                            [current_tool_call_chunk], tool_calls_stream
                        )
                        output = function_schema.parse_args(
                            self._tool_call(
                                tool_calls_stream, tool_call_ref, current_tool_call_id
                            )
                        )
                        yield output
                        if not tool_call_ref and not self._exhausted:
                            # Finish the group to allow advancing to the next one
                            # Output must be Iterable if parse_args above did not consume
                            assert isinstance(output, Iterable), output
                            # Consume stream via the output type so it can cache
                            consume(output)

                    except ValidationError as e:
                        assert current_tool_call_id is not None
                        raise ToolSchemaParseError(
                            output_message=self._state.current_message_snapshot,
                            tool_call_id=current_tool_call_id,
                            validation_error=e,
                        ) from e
            elif new_current_item := next(stream, None):
                current_item_ref.append(new_current_item)

    @property
    def usage_ref(self) -> list[Usage]:
        return self._state.usage_ref


class AsyncOutputStream(Generic[ItemT, OutputT]):
    """Async version of `OutputStream`."""

    def __init__(
        self,
        stream: AsyncIterator[ItemT],
        function_schemas: Iterable[AsyncFunctionSchema[OutputT]],
        parser: StreamParser[ItemT],
        state: StreamState[ItemT],
    ):
        self._stream = stream
        self._function_schemas = function_schemas
        self._parser = parser
        self._state = state

        self._iterator = self.__stream__()
        self._exhausted: bool = False

    async def __anext__(self) -> AsyncStreamedStr | OutputT:
        return await self._iterator.__anext__()

    async def __aiter__(self) -> AsyncIterator[AsyncStreamedStr | OutputT]:
        async for item in self._iterator:
            yield item

    async def _streamed_str(
        self, stream: AsyncIterator[ItemT], current_item_ref: list[ItemT]
    ) -> AsyncIterator[str]:
        async for item in stream:
            if content := self._parser.get_content(item):
                yield content
            if self._parser.is_tool_call(item):
                # TODO: Check if output types allow for early return
                assert not current_item_ref
                current_item_ref.append(item)
                return
        self._exhausted = True

    async def _tool_call(
        self,
        stream: AsyncIterator[FunctionCallChunk],
        current_tool_call_ref: list[FunctionCallChunk],
        current_tool_call_id: str,
    ) -> AsyncIterator[str]:
        async for item in stream:
            if item.id and item.id != current_tool_call_id:
                # TODO: Check if output types allow for early return
                assert not current_tool_call_ref
                current_tool_call_ref.append(item)
                return
            if item.args:
                yield item.args
        self._exhausted = True

    async def __stream__(self) -> AsyncIterator[AsyncStreamedStr | OutputT]:
        stream = aapply(self._state.update, self._stream)
        current_item_ref = [await anext(stream)]
        while current_item_ref:
            current_item = current_item_ref.pop()
            if self._parser.is_content(current_item):
                stream = achain(async_iter([current_item]), stream)
                streamed_str = AsyncStreamedStr(
                    self._streamed_str(stream, current_item_ref)
                )
                yield streamed_str
                if not current_item_ref and not self._exhausted:
                    # Finish the group to allow advancing to the next one
                    # Consume stream via AsyncStreamedStr so it can cache
                    await aconsume(streamed_str)
            elif self._parser.is_tool_call(current_item):
                tool_calls_stream: AsyncIterator[FunctionCallChunk] = (
                    tool_call_chunk
                    async for item in achain(async_iter([current_item]), stream)
                    for tool_call_chunk in self._parser.iter_tool_calls(item)
                )
                tool_call_ref = [await anext(tool_calls_stream)]
                while tool_call_ref:
                    current_tool_call_chunk = tool_call_ref.pop()
                    current_tool_call_id = current_tool_call_chunk.id
                    assert current_tool_call_id is not None
                    assert current_tool_call_chunk.name is not None
                    function_schema = select_function_schema(
                        self._function_schemas, current_tool_call_chunk.name
                    )
                    if function_schema is None:
                        assert current_tool_call_id is not None
                        raise UnknownToolError(
                            output_message=self._state.current_message_snapshot,
                            tool_call_id=current_tool_call_id,
                            tool_name=current_tool_call_chunk.name,
                        )
                    try:
                        tool_calls_stream = achain(
                            async_iter([current_tool_call_chunk]), tool_calls_stream
                        )
                        output = await function_schema.aparse_args(
                            self._tool_call(
                                tool_calls_stream, tool_call_ref, current_tool_call_id
                            )
                        )
                        yield output
                        if not tool_call_ref and not self._exhausted:
                            # Finish the group to allow advancing to the next one
                            # Output must be AsyncIterable if aparse_args above did not consume
                            assert isinstance(output, AsyncIterable), output
                            # Consume stream via the output type so it can cache
                            await aconsume(output)
                    except ValidationError as e:
                        assert current_tool_call_id is not None
                        raise ToolSchemaParseError(
                            output_message=self._state.current_message_snapshot,
                            tool_call_id=current_tool_call_id,
                            validation_error=e,
                        ) from e
            elif new_current_item := await anext(stream, None):
                current_item_ref.append(new_current_item)

    @property
    def usage_ref(self) -> list[Usage]:
        return self._state.usage_ref



from ._chat import Chat as Chat
from ._pydantic import ConfigDict as ConfigDict
from ._pydantic import with_config as with_config
from ._streamed_response import AsyncStreamedResponse as AsyncStreamedResponse
from ._streamed_response import StreamedResponse as StreamedResponse
from .chat_model.message import AnyMessage as AnyMessage
from .chat_model.message import AssistantMessage as AssistantMessage
from .chat_model.message import DocumentBytes as DocumentBytes
from .chat_model.message import FunctionResultMessage as FunctionResultMessage
from .chat_model.message import ImageBytes as ImageBytes
from .chat_model.message import ImageUrl as ImageUrl
from .chat_model.message import Placeholder as Placeholder
from .chat_model.message import SystemMessage as SystemMessage
from .chat_model.message import ToolResultMessage as ToolResultMessage
from .chat_model.message import UserMessage as UserMessage
from .chat_model.openai_chat_model import OpenaiChatModel as OpenaiChatModel
from .chatprompt import chatprompt as chatprompt
from .function_call import AsyncParallelFunctionCall as AsyncParallelFunctionCall
from .function_call import FunctionCall as FunctionCall
from .function_call import ParallelFunctionCall as ParallelFunctionCall
from .prompt_chain import prompt_chain as prompt_chain
from .prompt_function import prompt as prompt
from .streaming import AsyncStreamedStr as AsyncStreamedStr
from .streaming import StreamedStr as StreamedStr



from typing import Any, Generic, TypeVar, overload

from typing_extensions import deprecated

from magentic.chat_model.message import Message, Placeholder

T = TypeVar("T", bytes, str)
ImageContentT = TypeVar("ImageContentT")


@deprecated(
    "UserImageMessage is deprecated and will be removed in a future version."
    " Use UserMessage with ImageBytes and ImageUrl types instead."
)
class UserImageMessage(Message[ImageContentT], Generic[ImageContentT]):
    """A message containing an image sent by a user to an LLM chat model."""

    def __init__(self, content: ImageContentT, **data: Any):
        super().__init__(content=content, **data)

    @overload
    def format(
        self: "UserImageMessage[Placeholder[T]]", **kwargs: Any
    ) -> "UserImageMessage[T]": ...

    @overload
    def format(self: "UserImageMessage[T]", **kwargs: Any) -> "UserImageMessage[T]": ...

    def format(
        self: "UserImageMessage[Placeholder[T]] | UserImageMessage[T]",
        **kwargs: Any,
    ) -> "UserImageMessage[T]":
        """Format the message using the provided substitutions."""
        if isinstance(self.content, Placeholder):
            return UserImageMessage(self.content.format(**kwargs))
        return UserImageMessage(self.content)



import logging

import logfire_api

logger = logging.getLogger("magentic")
# Set default log level to WARNING so INFO logs must be explicitly enabled
if logger.level == logging.NOTSET:
    logger.setLevel(logging.WARNING)

logfire = logfire_api.Logfire(otel_scope="magentic")  # TODO: Pass version here too



import inspect
from typing import TYPE_CHECKING

import pytest
from typing_extensions import assert_type

from magentic.function_call import (
    AsyncParallelFunctionCall,
    FunctionCall,
    ParallelFunctionCall,
)
from magentic.streaming import async_iter

if TYPE_CHECKING:
    from collections.abc import Awaitable


def plus(a: int, b: int) -> int:
    return a + b


async def async_plus(a: int, b: int) -> int:
    return a + b


def plus_default_value(a: int, b: int = 3) -> int:
    return a + b


def return_hello() -> str:
    return "hello"


@pytest.mark.parametrize(
    ("left", "right", "equal"),
    [
        (FunctionCall(plus, a=1, b=2), FunctionCall(plus, a=1, b=2), True),
        (FunctionCall(plus, a=1, b=2), FunctionCall(plus, a=1, b=33), False),
        (
            FunctionCall(plus, a=1, b=2),
            FunctionCall(plus_default_value, a=1, b=2),
            False,
        ),
        (
            FunctionCall(plus_default_value, a=1),
            FunctionCall(plus_default_value, a=1, b=3),
            False,  # TODO: Should default values be considered? That would make this True.
        ),
        (
            FunctionCall(plus_default_value, a=1),
            FunctionCall(plus_default_value, a=1, b=44),
            False,
        ),
    ],
)
def test_function_call_eq(left, right, equal):
    assert (left == right) is equal, (left, right, equal)


@pytest.mark.parametrize(
    ("function_call", "arguments"),
    [
        (FunctionCall(plus, a=1, b=2), {"a": 1, "b": 2}),
        (FunctionCall(plus, 1, 2), {"a": 1, "b": 2}),
        (FunctionCall(plus_default_value, a=1), {"a": 1}),
    ],
)
def test_function_call_arguments(function_call, arguments):
    assert function_call.arguments == arguments


async def test_function_call_async_function():
    async def async_plus(a: int, b: int) -> int:
        return a + b

    function_call = FunctionCall(async_plus, a=1, b=2)
    result = function_call()
    assert inspect.isawaitable(result)
    assert await result == 3


def test_parallel_function_call_call():
    parallel_function_call: ParallelFunctionCall[int | str] = ParallelFunctionCall(
        [FunctionCall(plus, a=1, b=2), FunctionCall(return_hello)]
    )
    assert_type(parallel_function_call, ParallelFunctionCall[int | str])
    result = parallel_function_call()
    assert_type(result, tuple[int | str, ...])
    assert result == (3, "hello")


def test_parallel_function_call_iter():
    function_calls: list[FunctionCall[int | str]] = [
        FunctionCall(plus, a=1, b=2),
        FunctionCall(return_hello),
    ]
    parallel_function_call = ParallelFunctionCall(iter(function_calls))
    assert list(parallel_function_call) == function_calls
    assert list(parallel_function_call) == function_calls


async def test_async_parallel_function_call_call():
    function_calls: list[FunctionCall[int | Awaitable[int]]] = [
        FunctionCall(plus, a=1, b=2),
        FunctionCall(async_plus, a=3, b=4),
    ]
    async_parallel_function_call = AsyncParallelFunctionCall(async_iter(function_calls))
    assert_type(async_parallel_function_call, AsyncParallelFunctionCall[int])
    result = await async_parallel_function_call()
    assert_type(result, tuple[int, ...])
    assert result == (3, 7)


async def test_async_parallel_function_call_aiter():
    function_calls: list[FunctionCall[int | Awaitable[int]]] = [
        FunctionCall(plus, a=1, b=2),
        FunctionCall(async_plus, a=3, b=4),
    ]
    async_parallel_function_call = AsyncParallelFunctionCall(async_iter(function_calls))
    assert [x async for x in async_parallel_function_call] == function_calls
    assert [x async for x in async_parallel_function_call] == function_calls
